@@ -1,11 +1,10 @@
 from __future__ import annotations
 
 from dataclasses import dataclass
-from typing import Any, Dict, List
+from typing import Any, Dict, List, Optional, Tuple
 
 from qoala.lang.ehi import UnitModule
 from qoala.lang.program import QoalaProgram
-from qoala.runtime.schedule import BlockTask
 
 
 @dataclass
@@ -41,7 +40,6 @@ class ProgramInstance:
     program: QoalaProgram
     inputs: ProgramInput
     unit_module: UnitModule
-    block_tasks: List[BlockTask]
 
 
 @dataclass
@@ -55,3 +53,9 @@ class ProgramBatch:
 class BatchResult:
     batch_id: int
     results: List[ProgramResult]
+    timestamps: List[Optional[Tuple[float, float]]]  # start, end
+
+    @property
+    def durations(self) -> List[float]:
+        assert all(entry is not None for entry in self.timestamps)
+        return [end - start for (start, end) in self.timestamps]  # type: ignore
