@@ -1,8 +1,11 @@
 import logging
 from dataclasses import dataclass
-from typing import Dict, Optional
+from typing import Dict, List, Optional
+
+from netsquid.protocols import Protocol
 
 from qoala.lang.ehi import EhiNodeInfo, UnitModule
+from qoala.sim.events import SIGNAL_MEMORY_FREED
 from qoala.sim.process import QoalaProcess
 from qoala.sim.qdevice import QDevice
 from qoala.util.logging import LogManager
@@ -35,7 +38,7 @@ class VirtualLocation:
     virt_id: int
 
 
-class MemoryManager:
+class MemoryManager(Protocol):
     def __init__(
         self,
         node_name: str,
@@ -55,6 +58,8 @@ class MemoryManager:
             i: None for i in qdevice.get_all_qubit_ids()
         }  # phys ID -> virt location
 
+        self.add_signal(SIGNAL_MEMORY_FREED)
+
     def _get_free_comm_phys_id(self) -> int:
         for phys_id in self._qdevice.get_comm_qubit_ids():
             if self._physical_mapping[phys_id] is None:
@@ -81,6 +86,9 @@ class MemoryManager:
     def get_process(self, pid: int) -> QoalaProcess:
         return self._processes[pid]
 
+    def get_all_program_ids(self) -> List[int]:
+        return list(self._processes.keys())
+
     def allocate(self, pid: int, virt_id: int) -> int:
         vmap = self._process_mappings[pid]
         # Check if the virtual ID is in the unit module
@@ -118,19 +126,22 @@ class MemoryManager:
         vmap = self._process_mappings[pid]
         # Check if the virtual ID is in the unit module
         assert virt_id in vmap.unit_module.get_all_qubit_ids()
-        assert virt_id in self._process_mappings[pid].mapping
+        assert virt_id in vmap.mapping
 
-        phys_id = self._process_mappings[pid].mapping[virt_id]
+        phys_id = vmap.mapping[virt_id]
         if phys_id is None:
             raise AllocError
 
         # update mappings
         self._physical_mapping[phys_id] = None
-        self._process_mappings[pid].mapping[virt_id] = None
+        vmap.mapping[virt_id] = None
 
         # update netsquid memory
         self._qdevice.set_mem_pos_in_use(phys_id, False)
 
+        # send a signal for components that may be blocked on resources
+        self.send_signal(SIGNAL_MEMORY_FREED)
+
     def get_unmapped_non_comm_qubit(self, pid: int) -> int:
         """returns virt ID"""
         vp_map = self._process_mappings[pid].mapping
@@ -145,7 +156,10 @@ class MemoryManager:
         return min(free_ids)
 
     def phys_id_for(self, pid: int, virt_id: int) -> Optional[int]:
-        phys_id = self._process_mappings[pid].mapping[virt_id]
+        virt_mapping = self._process_mappings[pid]
+        if virt_id not in virt_mapping.mapping:
+            raise RuntimeError(f"virt ID {virt_id} not in Unit Module")
+        phys_id = virt_mapping.mapping[virt_id]
         return phys_id
 
     def virt_id_for(self, pid: int, phys_id: int) -> Optional[int]:
