@@ -27,7 +27,6 @@ from netsquid.components.instructions import Instruction as NsInstr
 from pydynaa import EventExpression
 from qoala.lang.routine import LocalRoutine
 from qoala.runtime.memory import ProgramMemory, RunningLocalRoutine
-from qoala.runtime.message import LrCallTuple, Message
 from qoala.runtime.sharedmem import MemAddr
 from qoala.sim.memmgr import NotAllocatedError
 from qoala.sim.process import QoalaProcess
@@ -99,17 +98,6 @@ class QnosProcessor:
         running_routine = RunningLocalRoutine(instance, input_addr, result_addr)
         process.qnos_mem.add_running_local_routine(running_routine)
 
-    def await_local_routine_call(
-        self, process: QoalaProcess
-    ) -> Generator[EventExpression, None, None]:
-        msg = yield from self._interface.receive_host_msg()
-        payload: LrCallTuple = msg.content
-        yield from self.assign_local_routine(
-            process, payload.routine_name, payload.input_addr, payload.result_addr
-        )
-        # Mock sending signal back to Host that subroutine has finished.
-        self._interface.send_host_msg(Message(None))
-
     def assign_local_routine(
         self,
         process: QoalaProcess,
@@ -152,7 +140,7 @@ class QnosProcessor:
         # TODO: handle program counter and jumping!!
         instr = subroutine.instructions[instr_idx]
         self._logger.debug(
-            f"{ns.sim_time()} interpreting instruction {instr} at line {instr_idx}"
+            f"{ns.sim_time()} interpreting instruction {instr_idx}: {instr}"
         )
 
         next_instr_idx: int
@@ -162,8 +150,10 @@ class QnosProcessor:
             or isinstance(instr, core.BranchUnaryInstruction)
             or isinstance(instr, core.BranchBinaryInstruction)
         ):
-            if new_line := (yield from self._interpret_branch_instr(pid, instr)):
-                next_instr_idx = new_line
+            if new_line_relative := (
+                yield from self._interpret_branch_instr(pid, instr)
+            ):
+                next_instr_idx = instr_idx + new_line_relative
             else:
                 next_instr_idx = instr_idx + 1
         else:
@@ -279,12 +269,7 @@ class QnosProcessor:
         result_addr = self._routine().result_addr
 
         value = qnos_mem.get_reg_value(instr.reg)
-        if value is None:
-            raise RuntimeError(f"value in register {instr.reg} is not defined")
-        self._logger.debug(
-            f"Storing value {value} from register {instr.reg} "
-            f"to array entry {instr.entry}"
-        )
+
         addr = instr.entry.address.address
         # Only allow NetQASM 2.0 input/output addresses
         assert isinstance(addr, str)
@@ -292,6 +277,10 @@ class QnosProcessor:
         entry = instr.entry.index
         assert isinstance(entry, Register)
         index = qnos_mem.get_reg_value(entry)
+        self._logger.debug(
+            f"Storing value {value} from register {instr.reg} "
+            f"to array entry {instr.entry} (index {entry} = {index})"
+        )
 
         # Simulate instruction duration.
         yield from self._interface.wait(self._latencies.qnos_instr_time)
@@ -357,20 +346,18 @@ class QnosProcessor:
         qnos_mem = self._prog_mem().qnos_mem
         a, b = None, None
         registers = []
+        condition = False
         if isinstance(instr, core.BranchUnaryInstruction):
             a = qnos_mem.get_reg_value(instr.reg)
             registers = [instr.reg]
+            condition = instr.check_condition(a)
         elif isinstance(instr, core.BranchBinaryInstruction):
             a = qnos_mem.get_reg_value(instr.reg0)
             b = qnos_mem.get_reg_value(instr.reg1)
             registers = [instr.reg0, instr.reg1]
-
-        if isinstance(instr, core.JmpInstruction):
-            condition = True
-        elif isinstance(instr, core.BranchUnaryInstruction):
-            condition = instr.check_condition(a)
-        elif isinstance(instr, core.BranchBinaryInstruction):
             condition = instr.check_condition(a, b)
+        elif isinstance(instr, core.JmpInstruction):
+            condition = True
 
         # Simulate instruction duration.
         yield from self._interface.wait(self._latencies.qnos_instr_time)
@@ -435,7 +422,9 @@ class QnosProcessor:
         elif isinstance(instr, core.MulInstruction):
             return a * b
         elif isinstance(instr, core.DivInstruction):
-            return int(a / b)
+            return a // b
+        elif isinstance(instr, core.RemInstruction):
+            return a % b
         else:
             raise ValueError(f"{instr} cannot be used as binary classical function")
 
@@ -473,7 +462,7 @@ class QnosProcessor:
         return None
 
     def _interpret_single_rotation_instr(
-        self, pid: int, instr: nv.RotXInstruction
+        self, pid: int, instr: core.RotationInstruction
     ) -> Generator[EventExpression, None, None]:
         raise NotImplementedError
 
@@ -523,7 +512,7 @@ class QnosProcessor:
         raise NotImplementedError
 
     def _interpret_two_qubit_instr(
-        self, pid: int, instr: core.SingleQubitInstruction
+        self, pid: int, instr: core.TwoQubitInstruction
     ) -> Generator[EventExpression, None, None]:
         raise NotImplementedError
 
@@ -592,7 +581,7 @@ class GenericProcessor(QnosProcessor):
         return None
 
     def _interpret_single_rotation_instr(
-        self, pid: int, instr: nv.RotXInstruction
+        self, pid: int, instr: core.RotationInstruction
     ) -> Generator[EventExpression, None, None]:
         if isinstance(instr, vanilla.RotXInstruction):
             yield from self._do_single_rotation(pid, instr, INSTR_ROT_X)
@@ -610,7 +599,7 @@ class GenericProcessor(QnosProcessor):
         raise UnsupportedNetqasmInstructionError
 
     def _interpret_two_qubit_instr(
-        self, pid: int, instr: core.SingleQubitInstruction
+        self, pid: int, instr: core.TwoQubitInstruction
     ) -> Generator[EventExpression, None, None]:
         qnos_mem = self._prog_mem().qnos_mem
         virt_id0 = qnos_mem.get_reg_value(instr.reg0)
@@ -638,9 +627,10 @@ class NVProcessor(QnosProcessor):
     def _interpret_init(
         self, pid: int, instr: core.InitInstruction
     ) -> Generator[EventExpression, None, None]:
+        memmgr = self._interface.memmgr
         qnos_mem = self._prog_mem().qnos_mem
         virt_id = qnos_mem.get_reg_value(instr.reg)
-        phys_id = self._interface.memmgr.phys_id_for(pid, virt_id)
+        phys_id = memmgr.phys_id_for(pid, virt_id)
         if phys_id is None:
             raise NotAllocatedError
         self._logger.debug(
@@ -648,6 +638,12 @@ class NVProcessor(QnosProcessor):
             f"{virt_id} (physical ID: {phys_id})"
         )
         commands = [QDeviceCommand(INSTR_INIT, [phys_id])]
+        if phys_id != 0 and memmgr.virt_id_for(pid, 0) is not None:
+            self._logger.info(
+                f"Initialization of virt qubit {virt_id} (phys ID {phys_id}) will "
+                "trash the existing state of phys qubit 0."
+            )
+            commands.append(QDeviceCommand(INSTR_INIT, [0]))
         yield from self.qdevice.execute_commands(commands)
         return None
 
@@ -717,8 +713,6 @@ class NVProcessor(QnosProcessor):
             f"placing the outcome in register {instr.creg}"
         )
 
-        memmgr = self._interface.memmgr
-
         if phys_id == 0:
             # Measuring a comm qubit. This can be done immediately.
             outcome = yield from self._measure_electron()
@@ -727,55 +721,16 @@ class NVProcessor(QnosProcessor):
         # else:
 
         # We want to measure a mem qubit.
-        # Move it to the comm qubit first.
-        # Check if comm qubit is already used by this process:
-        if memmgr.phys_id_for(pid, 0) is not None:
-            # Comm qubit is already allocated. Try to move it to a free mem qubit.
-            # mem_virt_id = memmgr.get_unmapped_non_comm_qubit(pid)
-            # memmgr.allocate(pid, mem_virt_id)
-            # mem_phys_id = memmgr.phys_id_for(pid, mem_virt_id)
-            # assert mem_phys_id is not None
+        # Move its state to the comm qubit.
+        yield from self._move_carbon_to_electron_for_measure(phys_id)
 
-            # Move (temporarily) state from comm qubit to mem qubit.
-            # yield from self._move_electron_to_carbon(mem_phys_id)
-
-            # Move state from qubit-to-measure to comm qubit.
-            yield from self._move_carbon_to_electron_for_measure(phys_id)
-            # memmgr.free(pid, virt_id)
-            # self._interface.signal_memory_freed()
-
-            # Measure comm qubit (containing state of qubit-to-measure).
-            outcome = yield from self._measure_electron()
-            qnos_mem.set_reg_value(instr.creg, outcome)
-
-            # Move temporarily-moved state back to comm qubit.
-            # yield from self._move_carbon_to_electron(mem_phys_id)
-
-            # Free mem qubit that was temporarily used.
-            # memmgr.free(pid, mem_virt_id)
-            # self._interface.signal_memory_freed()
-        else:  # comm qubit not in use.
-            # Allocate comm qubit.
-            # memmgr.allocate(pid, 0)
-
-            # Move state-to-measure to comm qubit.
-            yield from self._move_carbon_to_electron_for_measure(phys_id)
-
-            # Free qubit that contained state-to-measure.
-            # memmgr.free(pid, virt_id)
-            # self._interface.signal_memory_freed()
-
-            # Measure comm qubit.
-            outcome = yield from self._measure_electron()
-            qnos_mem.set_reg_value(instr.creg, outcome)
-
-            # Free comm qubit.
-            # memmgr.free(pid, 0)
-            # self._interface.signal_memory_freed()
+        # Measure comm qubit (containing state of qubit to measure).
+        outcome = yield from self._measure_electron()
+        qnos_mem.set_reg_value(instr.creg, outcome)
         return None
 
     def _interpret_single_rotation_instr(
-        self, pid: int, instr: nv.RotXInstruction
+        self, pid: int, instr: core.RotationInstruction
     ) -> Generator[EventExpression, None, None]:
         if isinstance(instr, nv.RotXInstruction):
             yield from self._do_single_rotation(pid, instr, INSTR_ROT_X)
