@@ -4,7 +4,7 @@ import netsquid as ns
 from netsquid.nodes import Node
 
 from pydynaa import EventExpression
-from qoala.lang.ehi import EhiBuilder, UnitModule
+from qoala.lang.ehi import EhiBuilder, EhiNetworkInfo, UnitModule
 from qoala.lang.program import ProgramMeta, QoalaProgram
 from qoala.lang.request import (
     CallbackType,
@@ -15,7 +15,6 @@ from qoala.lang.request import (
     RequestVirtIdMapping,
 )
 from qoala.lang.routine import LocalRoutine
-from qoala.runtime.environment import LocalEnvironment, NetworkInfo
 from qoala.runtime.lhi import LhiLinkInfo, LhiTopologyBuilder
 from qoala.runtime.memory import ProgramMemory
 from qoala.runtime.message import Message, RrCallTuple
@@ -24,12 +23,14 @@ from qoala.runtime.sharedmem import MemAddr
 from qoala.sim.build import build_qprocessor_from_topology
 from qoala.sim.entdist.entdist import EntDist, EntDistRequest
 from qoala.sim.entdist.entdistcomp import EntDistComponent
+from qoala.sim.eprsocket import EprSocket
 from qoala.sim.memmgr import MemoryManager
 from qoala.sim.netstack import NetstackInterface, NetstackLatencies
 from qoala.sim.netstack.netstack import Netstack
 from qoala.sim.netstack.netstackcomp import NetstackComponent
 from qoala.sim.process import QoalaProcess
 from qoala.sim.qdevice import QDevice
+from qoala.util.logging import LogManager
 from qoala.util.math import B00_DENS, has_multi_state
 
 
@@ -37,11 +38,11 @@ class MockNetstackInterface(NetstackInterface):
     def __init__(
         self,
         comp: NetstackComponent,
-        local_env: LocalEnvironment,
+        ehi_network: EhiNetworkInfo,
         qdevice: QDevice,
         requests: List[EntDistRequest],
     ) -> None:
-        super().__init__(comp, local_env, qdevice, None)
+        super().__init__(comp, ehi_network, qdevice, None)
         self._requests = requests
 
 
@@ -74,10 +75,18 @@ def create_alice_bob_qdevices(
 
 
 def create_request(
-    node1_id: int, node2_id: int, local_qubit_id: int = 0
+    node1_id: int,
+    node2_id: int,
+    local_qubit_id: int = 0,
+    lpid: Optional[int] = 0,
+    rpid: Optional[int] = 0,
 ) -> EntDistRequest:
     return EntDistRequest(
-        local_node_id=node1_id, remote_node_id=node2_id, local_qubit_id=local_qubit_id
+        local_node_id=node1_id,
+        remote_node_id=node2_id,
+        local_qubit_id=local_qubit_id,
+        local_pid=lpid,
+        remote_pid=rpid,
     )
 
 
@@ -86,14 +95,16 @@ def setup_components() -> Tuple[
 ]:
     alice, bob = create_alice_bob_qdevices(num_qubits=3)
 
-    env = NetworkInfo.with_nodes(
+    ehi_network = EhiNetworkInfo.only_nodes(
         {alice.node.ID: alice.node.name, bob.node.ID: bob.node.name}
     )
-    alice_comp = NetstackComponent(alice.node, env)
-    bob_comp = NetstackComponent(bob.node, env)
-    entdist_comp = EntDistComponent(env)
+    alice_comp = NetstackComponent(alice.node, ehi_network)
+    bob_comp = NetstackComponent(bob.node, ehi_network)
+    entdist_comp = EntDistComponent(ehi_network)
 
-    entdist = EntDist(nodes=[alice.node, bob.node], network_info=env, comp=entdist_comp)
+    entdist = EntDist(
+        nodes=[alice.node, bob.node], ehi_network=ehi_network, comp=entdist_comp
+    )
 
     alice_comp.entdist_out_port.connect(entdist_comp.node_in_port("alice"))
     alice_comp.entdist_in_port.connect(entdist_comp.node_out_port("alice"))
@@ -110,36 +121,37 @@ def test_single_pair_only_netstack_interface():
     class AliceNetstackInterface(MockNetstackInterface):
         def run(self) -> Generator[EventExpression, None, None]:
             yield from self.wait(500)
-            self.send_entdist_msg(Message(self._requests[0]))
+            self.send_entdist_msg(Message(0, 0, self._requests[0]))
 
     class BobNetstackInterface(MockNetstackInterface):
         def run(self) -> Generator[EventExpression, None, None]:
             yield from self.wait(800)
-            self.send_entdist_msg(Message(self._requests[0]))
+            self.send_entdist_msg(Message(0, 0, self._requests[0]))
 
     alice_comp, alice_qdevice, bob_comp, bob_qdevice, entdist = setup_components()
-    env: NetworkInfo = entdist._network_info
+    ehi_network: EhiNetworkInfo = entdist._ehi_network
     alice_id = alice_comp.node.ID
     bob_id = bob_comp.node.ID
 
-    request_alice = create_request(alice_id, bob_id)
-    request_bob = create_request(bob_id, alice_id)
+    request_alice = create_request(alice_id, bob_id, 0, 0, 0)
+    request_bob = create_request(bob_id, alice_id, 0, 0, 0)
 
     alice_intf = AliceNetstackInterface(
-        alice_comp,
-        LocalEnvironment(env, alice_id),
-        alice_qdevice,
-        requests=[request_alice],
+        alice_comp, ehi_network, alice_qdevice, requests=[request_alice]
     )
     bob_intf = BobNetstackInterface(
-        bob_comp, LocalEnvironment(env, bob_id), bob_qdevice, requests=[request_bob]
+        bob_comp, ehi_network, bob_qdevice, requests=[request_bob]
     )
+    ns.sim_reset()
+    assert ns.sim_time() == 0
 
     alice_intf.start()
     bob_intf.start()
     entdist.start()
     ns.sim_run()
 
+    assert ns.sim_time() == 1800  # 1000 + max(800, 500)
+
     alice_qubit = alice_qdevice.get_local_qubit(0)
     bob_qubit = bob_qdevice.get_local_qubit(0)
     assert has_multi_state([alice_qubit, bob_qubit], B00_DENS)
@@ -150,51 +162,60 @@ def test_multiple_pairs_only_netstack_interface():
         def run(self) -> Generator[EventExpression, None, None]:
             for request in self._requests:
                 yield from self.wait(500)
-                self.send_entdist_msg(Message(request))
+                self.send_entdist_msg(Message(0, 0, request))
 
     class BobNetstackInterface(MockNetstackInterface):
         def run(self) -> Generator[EventExpression, None, None]:
             for request in self._requests:
                 yield from self.wait(500)
-                self.send_entdist_msg(Message(request))
+                self.send_entdist_msg(Message(0, 0, request))
 
     alice_comp, alice_qdevice, bob_comp, bob_qdevice, entdist = setup_components()
-    env: NetworkInfo = entdist._network_info
+    ehi_network: EhiNetworkInfo = entdist._ehi_network
     alice_id = alice_comp.node.ID
     bob_id = bob_comp.node.ID
 
     requests_alice = [
-        create_request(alice_id, bob_id, 0),
-        create_request(alice_id, bob_id, 1),
-        create_request(alice_id, bob_id, 2),
+        create_request(alice_id, bob_id, 0, 0, 0),
+        create_request(alice_id, bob_id, 1, 0, 0),
+        create_request(alice_id, bob_id, 2, 0, 0),
     ]
     requests_bob = [
-        create_request(bob_id, alice_id, 1),
-        create_request(bob_id, alice_id, 2),
-        create_request(bob_id, alice_id, 0),
+        create_request(bob_id, alice_id, 1, 0, 0),
+        create_request(bob_id, alice_id, 2, 0, 0),
+        create_request(bob_id, alice_id, 0, 0, 0),
     ]
 
     alice_intf = AliceNetstackInterface(
-        alice_comp,
-        LocalEnvironment(env, alice_id),
-        alice_qdevice,
-        requests=requests_alice,
+        alice_comp, ehi_network, alice_qdevice, requests=requests_alice
     )
     bob_intf = BobNetstackInterface(
-        bob_comp, LocalEnvironment(env, bob_id), bob_qdevice, requests=requests_bob
+        bob_comp, ehi_network, bob_qdevice, requests=requests_bob
     )
 
+    ns.sim_reset()
+    assert ns.sim_time() == 0
     alice_intf.start()
     bob_intf.start()
     entdist.start()
     ns.sim_run()
 
+    # 500 + 1000 * 3 since epr pairs are created sequentially
+    assert ns.sim_time() == 3500
+
     alice_q0, alice_q1, alice_q2 = [alice_qdevice.get_local_qubit(i) for i in range(3)]
     bob_q0, bob_q1, bob_q2 = [bob_qdevice.get_local_qubit(i) for i in range(3)]
     assert has_multi_state([alice_q0, bob_q1], B00_DENS)
     assert has_multi_state([alice_q1, bob_q2], B00_DENS)
     assert has_multi_state([alice_q2, bob_q0], B00_DENS)
 
+    assert not has_multi_state([alice_q0, bob_q0], B00_DENS)
+    assert not has_multi_state([alice_q1, bob_q1], B00_DENS)
+    assert not has_multi_state([alice_q2, bob_q2], B00_DENS)
+    assert not has_multi_state([alice_q0, bob_q2], B00_DENS)
+    assert not has_multi_state([alice_q1, bob_q0], B00_DENS)
+    assert not has_multi_state([alice_q2, bob_q1], B00_DENS)
+
 
 def setup_components_full_netstack(
     num_qubits: int,
@@ -207,19 +228,19 @@ def setup_components_full_netstack(
         num_qubits=num_qubits, alice_id=alice_id, bob_id=bob_id
     )
 
-    env = NetworkInfo.with_nodes(
+    ehi_network = EhiNetworkInfo.only_nodes(
         {
             alice_qdevice.node.ID: alice_qdevice.node.name,
             bob_qdevice.node.ID: bob_qdevice.node.name,
         }
     )
-    alice_comp = NetstackComponent(alice_qdevice.node, env)
-    bob_comp = NetstackComponent(bob_qdevice.node, env)
-    entdist_comp = EntDistComponent(env)
+    alice_comp = NetstackComponent(alice_qdevice.node, ehi_network)
+    bob_comp = NetstackComponent(bob_qdevice.node, ehi_network)
+    entdist_comp = EntDistComponent(ehi_network)
 
     entdist = EntDist(
         nodes=[alice_qdevice.node, bob_qdevice.node],
-        network_info=env,
+        ehi_network=ehi_network,
         comp=entdist_comp,
     )
 
@@ -233,14 +254,14 @@ def setup_components_full_netstack(
 
     alice_netstack = alice_netstack_cls(
         comp=alice_comp,
-        local_env=LocalEnvironment(env, alice_qdevice.node.ID),
+        ehi_network=ehi_network,
         memmgr=MemoryManager("alice", alice_qdevice),
         qdevice=alice_qdevice,
         latencies=NetstackLatencies.all_zero(),
     )
     bob_netstack = bob_netstack_cls(
         comp=bob_comp,
-        local_env=LocalEnvironment(env, bob_qdevice.node.ID),
+        ehi_network=ehi_network,
         memmgr=MemoryManager("bob", bob_qdevice),
         qdevice=bob_qdevice,
         latencies=NetstackLatencies.all_zero(),
@@ -253,26 +274,30 @@ def test_single_pair_full_netstack():
     alice_id = 0
     bob_id = 1
 
-    request_alice = create_request(alice_id, bob_id)
-    request_bob = create_request(bob_id, alice_id)
+    request_alice = create_request(alice_id, bob_id, 0, 0, 0)
+    request_bob = create_request(bob_id, alice_id, 0, 0, 0)
 
     class AliceNetstack(Netstack):
         def run(self) -> Generator[EventExpression, None, None]:
-            yield from self.processor.execute_entdist_request(request_alice)
+            yield from self.processor._execute_entdist_request(request_alice)
 
     class BobNetstack(Netstack):
         def run(self) -> Generator[EventExpression, None, None]:
-            yield from self.processor.execute_entdist_request(request_bob)
+            yield from self.processor._execute_entdist_request(request_bob)
 
     alice_netstack, bob_netstack, entdist = setup_components_full_netstack(
         1, alice_id, bob_id, AliceNetstack, BobNetstack
     )
 
+    ns.sim_reset()
+    assert ns.sim_time() == 0
     alice_netstack.start()
     bob_netstack.start()
     entdist.start()
     ns.sim_run()
 
+    assert ns.sim_time() == 1000
+
     alice_qubit = alice_netstack.qdevice.get_local_qubit(0)
     bob_qubit = bob_netstack.qdevice.get_local_qubit(0)
     assert has_multi_state([alice_qubit, bob_qubit], B00_DENS)
@@ -285,23 +310,23 @@ def test_multiple_pairs_full_netstack():
     bob_id = 1
 
     requests_alice = [
-        create_request(alice_id, bob_id, 0),
-        create_request(alice_id, bob_id, 1),
+        create_request(alice_id, bob_id, 0, 0, 0),
+        create_request(alice_id, bob_id, 1, 1, 1),
     ]
     requests_bob = [
-        create_request(bob_id, alice_id, 0),
-        create_request(bob_id, alice_id, 2),
+        create_request(bob_id, alice_id, 0, 0, 0),
+        create_request(bob_id, alice_id, 2, 1, 1),
     ]
 
     class AliceNetstack(Netstack):
         def run(self) -> Generator[EventExpression, None, None]:
             for request in requests_alice:
-                yield from self.processor.execute_entdist_request(request)
+                yield from self.processor._execute_entdist_request(request)
 
     class BobNetstack(Netstack):
         def run(self) -> Generator[EventExpression, None, None]:
             for request in requests_bob:
-                yield from self.processor.execute_entdist_request(request)
+                yield from self.processor._execute_entdist_request(request)
 
     alice_netstack, bob_netstack, entdist = setup_components_full_netstack(
         3, alice_id, bob_id, AliceNetstack, BobNetstack
@@ -346,7 +371,7 @@ def simple_req_routine(
     virt_ids: RequestVirtIdMapping,
     typ: EprType,
     role: EprRole,
-) -> QoalaRequest:
+) -> RequestRoutine:
     return RequestRoutine(
         name="req1",
         request=create_simple_request(remote_id, num_pairs, virt_ids, typ, role),
@@ -360,11 +385,14 @@ def create_process(
     num_qubits: int,
     routines: Optional[Dict[str, LocalRoutine]] = None,
     req_routines: Optional[Dict[str, RequestRoutine]] = None,
+    epr_sockets: Optional[Dict[int, EprSocket]] = None,
 ) -> QoalaProcess:
     if routines is None:
         routines = {}
     if req_routines is None:
         req_routines = {}
+    if epr_sockets is None:
+        epr_sockets = {}
     program = QoalaProgram(
         blocks=[],
         local_routines=routines,
@@ -379,7 +407,6 @@ def create_process(
         program=program,
         inputs=ProgramInput({}),
         unit_module=unit_module,
-        block_tasks=[],
     )
     mem = ProgramMemory(pid=0)
 
@@ -387,7 +414,7 @@ def create_process(
         prog_instance=instance,
         prog_memory=mem,
         csockets={},
-        epr_sockets=program.meta.epr_sockets,
+        epr_sockets=epr_sockets,
         result=ProgramResult(values={}),
     )
     return process
@@ -407,10 +434,13 @@ def test_single_pair_qoala_ck_request_only_alice():
     )
 
     requests_bob = [
-        create_request(bob_id, alice_id, 0),
-        create_request(bob_id, alice_id, 1),
+        create_request(bob_id, alice_id, 0, 0, 0),
+        create_request(bob_id, alice_id, 1, 0, 0),
     ]
-    process_alice = create_process(num_qubits, req_routines={"req1": routine_alice})
+    epr_socket = EprSocket(0, bob_id, 0, 0, 1.0)
+    process_alice = create_process(
+        num_qubits, req_routines={"req1": routine_alice}, epr_sockets={0: epr_socket}
+    )
 
     class AliceNetstack(Netstack):
         def run(self) -> Generator[EventExpression, None, None]:
@@ -420,7 +450,7 @@ def test_single_pair_qoala_ck_request_only_alice():
     class BobNetstack(Netstack):
         def run(self) -> Generator[EventExpression, None, None]:
             for request in requests_bob:
-                yield from self.processor.execute_entdist_request(request)
+                yield from self.processor._execute_entdist_request(request)
 
     alice_netstack, bob_netstack, entdist = setup_components_full_netstack(
         num_qubits, alice_id, bob_id, AliceNetstack, BobNetstack
@@ -460,8 +490,16 @@ def test_single_pair_qoala_ck_request():
         role=EprRole.RECEIVE,
     )
 
-    process_alice = create_process(num_qubits, req_routines={"req1": routine_alice})
-    process_bob = create_process(num_qubits, req_routines={"req1": routine_bob})
+    epr_socket_alice = EprSocket(0, bob_id, 0, 0, 1.0)
+    process_alice = create_process(
+        num_qubits,
+        req_routines={"req1": routine_alice},
+        epr_sockets={0: epr_socket_alice},
+    )
+    epr_socket_bob = EprSocket(0, alice_id, 0, 0, 1.0)
+    process_bob = create_process(
+        num_qubits, req_routines={"req1": routine_bob}, epr_sockets={0: epr_socket_bob}
+    )
 
     class AliceNetstack(Netstack):
         def run(self) -> Generator[EventExpression, None, None]:
@@ -512,8 +550,16 @@ def test_single_pair_qoala_md_request_different_virt_ids():
         role=EprRole.RECEIVE,
     )
 
-    process_alice = create_process(num_qubits, req_routines={"req1": routine_alice})
-    process_bob = create_process(num_qubits, req_routines={"req1": routine_bob})
+    epr_socket_alice = EprSocket(0, bob_id, 0, 0, 1.0)
+    process_alice = create_process(
+        num_qubits,
+        req_routines={"req1": routine_alice},
+        epr_sockets={0: epr_socket_alice},
+    )
+    epr_socket_bob = EprSocket(0, alice_id, 0, 0, 1.0)
+    process_bob = create_process(
+        num_qubits, req_routines={"req1": routine_bob}, epr_sockets={0: epr_socket_bob}
+    )
 
     class AliceNetstack(Netstack):
         def run(self) -> Generator[EventExpression, None, None]:
@@ -585,8 +631,16 @@ def test_single_pair_qoala_md_request_same_virt_ids():
         role=EprRole.RECEIVE,
     )
 
-    process_alice = create_process(num_qubits, req_routines={"req1": routine_alice})
-    process_bob = create_process(num_qubits, req_routines={"req1": routine_bob})
+    epr_socket_alice = EprSocket(0, bob_id, 0, 0, 1.0)
+    process_alice = create_process(
+        num_qubits,
+        req_routines={"req1": routine_alice},
+        epr_sockets={0: epr_socket_alice},
+    )
+    epr_socket_bob = EprSocket(0, alice_id, 0, 0, 1.0)
+    process_bob = create_process(
+        num_qubits, req_routines={"req1": routine_bob}, epr_sockets={0: epr_socket_bob}
+    )
 
     class AliceNetstack(Netstack):
         def run(self) -> Generator[EventExpression, None, None]:
@@ -642,11 +696,12 @@ def test_single_pair_qoala_md_request_same_virt_ids():
 
 
 if __name__ == "__main__":
-    test_single_pair_only_netstack_interface()
+    # test_single_pair_only_netstack_interface()
+    LogManager.set_log_level("INFO")
     test_multiple_pairs_only_netstack_interface()
-    test_single_pair_full_netstack()
-    test_multiple_pairs_full_netstack()
-    test_single_pair_qoala_ck_request_only_alice()
-    test_single_pair_qoala_ck_request()
-    test_single_pair_qoala_md_request_different_virt_ids()
-    test_single_pair_qoala_md_request_same_virt_ids()
+    # test_single_pair_full_netstack()
+    # test_multiple_pairs_full_netstack()
+    # test_single_pair_qoala_ck_request_only_alice()
+    # test_single_pair_qoala_ck_request()
+    # test_single_pair_qoala_md_request_different_virt_ids()
+    # test_single_pair_qoala_md_request_same_virt_ids()
