@@ -1,10 +1,10 @@
 from __future__ import annotations
 
 from dataclasses import dataclass
-from typing import Any, Dict, Generator, List, Optional
+from typing import Any, Dict, Generator, List, Optional, Tuple
 
 import netsquid as ns
-from netqasm.lang.parsing import parse_text_subroutine
+import pytest
 from netqasm.lang.subroutine import Subroutine
 
 from pydynaa import EventExpression
@@ -16,7 +16,10 @@ from qoala.lang.hostlang import (
     BasicBlockType,
     BitConditionalMultiplyConstantCValueOp,
     ClassicalIqoalaOp,
+    IqoalaSingleton,
+    IqoalaTuple,
     IqoalaVector,
+    IqoalaVectorElement,
     MultiplyConstantCValueOp,
     ReceiveCMsgOp,
     ReturnResultOp,
@@ -35,7 +38,7 @@ from qoala.lang.request import (
 )
 from qoala.lang.routine import LocalRoutine, RoutineMetadata
 from qoala.runtime.memory import ProgramMemory
-from qoala.runtime.message import LrCallTuple, Message, RrCallTuple
+from qoala.runtime.message import Message
 from qoala.runtime.program import ProgramInput, ProgramInstance, ProgramResult
 from qoala.runtime.sharedmem import SharedMemory
 from qoala.sim.host.csocket import ClassicalSocket
@@ -44,13 +47,13 @@ from qoala.sim.host.hostprocessor import HostProcessor
 from qoala.sim.process import QoalaProcess
 from qoala.util.tests import netsquid_run, yield_from
 
-MOCK_MESSAGE = Message(content=42)
+MOCK_MESSAGE = Message(src_pid=0, dst_pid=0, content=42)
 MOCK_QNOS_RET_REG = "R0"
 MOCK_QNOS_RET_VALUE = 7
 MOCK_NETSTACK_RET_VALUE = 22
 
 
-@dataclass(eq=True, frozen=True)
+@dataclass(frozen=True)
 class InterfaceEvent:
     peer: str
     msg: Message
@@ -62,6 +65,11 @@ class MockHostInterface(HostInterface):
         self.recv_events: List[InterfaceEvent] = []
 
         self.shared_mem = shared_mem
+        self._program_instance_jumps: Dict[int, int] = {}  # pid => block name
+
+    @property
+    def program_instance_jumps(self) -> Dict[int, int]:
+        return self._program_instance_jumps
 
     def send_peer_msg(self, peer: str, msg: Message) -> None:
         self.send_events.append(InterfaceEvent(peer, msg))
@@ -74,6 +82,9 @@ class MockHostInterface(HostInterface):
     def send_qnos_msg(self, msg: Message) -> None:
         self.send_events.append(InterfaceEvent("qnos", msg))
 
+    def get_available_messages(self, peer: str) -> List[Tuple[int, int]]:
+        return []
+
     def receive_qnos_msg(self) -> Generator[EventExpression, None, Message]:
         self.recv_events.append(InterfaceEvent("qnos", MOCK_MESSAGE))
         if self.shared_mem is not None:
@@ -83,14 +94,6 @@ class MockHostInterface(HostInterface):
         return MOCK_MESSAGE
         yield  # to make it behave as a generator
 
-    def send_netstack_msg(self, msg: Message) -> None:
-        self.send_events.append(InterfaceEvent("netstack", msg))
-
-    def receive_netstack_msg(self) -> Generator[EventExpression, None, Message]:
-        self.recv_events.append(InterfaceEvent("netstack", MOCK_MESSAGE))
-        return MOCK_MESSAGE
-        yield  # to make it behave as a generator
-
     @property
     def name(self) -> str:
         return "mock"
@@ -104,10 +107,7 @@ def create_program(
 ) -> QoalaProgram:
     if instrs is None:
         instrs = []
-    if subroutines is None:
-        subroutines = {}
-    if requests is None:
-        requests = {}
+
     if meta is None:
         meta = ProgramMeta.empty("prog")
     # TODO: split into proper blocks
@@ -140,7 +140,6 @@ def create_process(
         program=program,
         inputs=prog_input,
         unit_module=UnitModule.from_full_ehi(mock_ehi),
-        block_tasks=[],
     )
 
     mem = ProgramMemory(pid=0)
@@ -148,7 +147,7 @@ def create_process(
         prog_instance=instance,
         prog_memory=mem,
         csockets={
-            id: ClassicalSocket(interface, name)
+            id: ClassicalSocket(interface, name, 0, 0)
             for (id, name) in program.meta.csockets.items()
         },
         epr_sockets=program.meta.epr_sockets,
@@ -167,15 +166,19 @@ def test_initialize():
 
     processor.initialize(process)
     host_mem = process.prog_memory.host_mem
+
     assert host_mem.read("x") == 1
     assert host_mem.read("theta") == 3.14
     assert host_mem.read("name") == "alice"
 
+    with pytest.raises(KeyError):
+        host_mem.read_vec("x")
+
 
 def test_assign_cvalue():
     interface = MockHostInterface()
     processor = HostProcessor(interface, HostLatencies.all_zero())
-    program = create_program(instrs=[AssignCValueOp("x", 3)])
+    program = create_program(instrs=[AssignCValueOp(IqoalaSingleton("x"), 3)])
     process = create_process(program, interface)
     processor.initialize(process)
 
@@ -189,7 +192,7 @@ def test_assign_cvalue_with_latencies():
     interface = MockHostInterface()
     latencies = HostLatencies(host_instr_time=500)
     processor = HostProcessor(interface, latencies)
-    program = create_program(instrs=[AssignCValueOp("x", 3)])
+    program = create_program(instrs=[AssignCValueOp(IqoalaSingleton("x"), 3)])
     process = create_process(program, interface)
     processor.initialize(process)
 
@@ -204,12 +207,15 @@ def test_send_msg():
     processor = HostProcessor(interface, HostLatencies.all_zero())
     meta = ProgramMeta.empty("alice")
     meta.csockets = {0: "bob"}
-    program = create_program(instrs=[SendCMsgOp("bob", "msg")], meta=meta)
+
+    program = create_program(
+        instrs=[SendCMsgOp(IqoalaSingleton("bob"), IqoalaSingleton("msg"))], meta=meta
+    )
     process = create_process(program, interface, inputs={"bob": 0, "msg": 12})
     processor.initialize(process)
 
     yield_from(processor.assign_instr_index(process, 0))
-    assert interface.send_events[0] == InterfaceEvent("bob", Message(content=12))
+    assert interface.send_events[0] == InterfaceEvent("bob", Message(0, 0, content=12))
 
 
 def test_recv_msg():
@@ -217,7 +223,10 @@ def test_recv_msg():
     processor = HostProcessor(interface, HostLatencies.all_zero())
     meta = ProgramMeta.empty("alice")
     meta.csockets = {0: "bob"}
-    program = create_program(instrs=[ReceiveCMsgOp("bob", "msg")], meta=meta)
+    program = create_program(
+        instrs=[ReceiveCMsgOp(IqoalaSingleton("bob"), IqoalaSingleton("msg"))],
+        meta=meta,
+    )
     process = create_process(program, interface, inputs={"bob": 0})
     processor.initialize(process)
 
@@ -234,7 +243,10 @@ def test_recv_msg_with_latencies():
     processor = HostProcessor(interface, latencies)
     meta = ProgramMeta.empty("alice")
     meta.csockets = {0: "bob"}
-    program = create_program(instrs=[ReceiveCMsgOp("bob", "msg")], meta=meta)
+    program = create_program(
+        instrs=[ReceiveCMsgOp(IqoalaSingleton("bob"), IqoalaSingleton("msg"))],
+        meta=meta,
+    )
     process = create_process(program, interface, inputs={"bob": 0})
     processor.initialize(process)
 
@@ -250,9 +262,11 @@ def test_add_cvalue():
     processor = HostProcessor(interface, HostLatencies.all_zero())
     program = create_program(
         instrs=[
-            AssignCValueOp("a", 2),
-            AssignCValueOp("b", 3),
-            AddCValueOp("sum", "a", "b"),
+            AssignCValueOp(IqoalaSingleton("a"), 2),
+            AssignCValueOp(IqoalaSingleton("b"), 3),
+            AddCValueOp(
+                IqoalaSingleton("sum"), IqoalaSingleton("a"), IqoalaSingleton("b")
+            ),
         ]
     )
     process = create_process(program, interface)
@@ -271,9 +285,11 @@ def test_add_cvalue_with_latencies():
     processor = HostProcessor(interface, HostLatencies(host_instr_time=1200))
     program = create_program(
         instrs=[
-            AssignCValueOp("a", 2),
-            AssignCValueOp("b", 3),
-            AddCValueOp("sum", "a", "b"),
+            AssignCValueOp(IqoalaSingleton("a"), 2),
+            AssignCValueOp(IqoalaSingleton("b"), 3),
+            AddCValueOp(
+                IqoalaSingleton("sum"), IqoalaSingleton("a"), IqoalaSingleton("b")
+            ),
         ]
     )
     process = create_process(program, interface)
@@ -291,7 +307,12 @@ def test_multiply_const():
     interface = MockHostInterface()
     processor = HostProcessor(interface, HostLatencies.all_zero())
     program = create_program(
-        instrs=[AssignCValueOp("a", 4), MultiplyConstantCValueOp("result", "a", -1)]
+        instrs=[
+            AssignCValueOp(IqoalaSingleton("a"), 4),
+            MultiplyConstantCValueOp(
+                IqoalaSingleton("result"), IqoalaSingleton("a"), -1
+            ),
+        ]
     )
     process = create_process(program, interface)
     processor.initialize(process)
@@ -307,12 +328,22 @@ def test_bit_cond_mult():
     processor = HostProcessor(interface, HostLatencies.all_zero())
     program = create_program(
         instrs=[
-            AssignCValueOp("var1", 4),
-            AssignCValueOp("var2", 7),
-            AssignCValueOp("cond1", 0),
-            AssignCValueOp("cond2", 1),
-            BitConditionalMultiplyConstantCValueOp("result1", "var1", "cond1", -1),
-            BitConditionalMultiplyConstantCValueOp("result2", "var2", "cond2", -1),
+            AssignCValueOp(IqoalaSingleton("var1"), 4),
+            AssignCValueOp(IqoalaSingleton("var2"), 7),
+            AssignCValueOp(IqoalaSingleton("cond1"), 0),
+            AssignCValueOp(IqoalaSingleton("cond2"), 1),
+            BitConditionalMultiplyConstantCValueOp(
+                IqoalaSingleton("result1"),
+                IqoalaSingleton("var1"),
+                IqoalaSingleton("cond1"),
+                -1,
+            ),
+            BitConditionalMultiplyConstantCValueOp(
+                IqoalaSingleton("result2"),
+                IqoalaSingleton("var2"),
+                IqoalaSingleton("cond2"),
+                -1,
+            ),
         ]
     )
     process = create_process(program, interface)
@@ -325,81 +356,55 @@ def test_bit_cond_mult():
     assert process.prog_memory.host_mem.read("result2") == -7
 
 
-def test_run_subroutine():
+def test_prepare_lr_call():
     interface = MockHostInterface()
-    processor = HostProcessor(interface, HostLatencies.all_zero(), asynchronous=True)
+    processor = HostProcessor(interface, HostLatencies.all_zero())
 
     subrt = Subroutine()
     metadata = RoutineMetadata.use_none()
-    routine = LocalRoutine("subrt1", subrt, return_vars=[], metadata=metadata)
-
-    program = create_program(
-        instrs=[RunSubroutineOp(None, IqoalaVector([]), "subrt1")],
-        subroutines={"subrt1": routine},
-    )
-    process = create_process(program, interface)
-    processor.initialize(process)
-
-    for i in range(len(program.instructions)):
-        yield_from(processor.assign_instr_index(process, i))
-
-    # Host processor should have communicated with the qnos processor.
-    send_event = interface.send_events[0]
-    assert send_event.peer == "qnos"
-    assert isinstance(send_event.msg, Message)
-    assert isinstance(send_event.msg.content, LrCallTuple)
-    assert send_event.msg.content.routine_name == "subrt1"
-    assert interface.recv_events[0] == InterfaceEvent("qnos", MOCK_MESSAGE)
-
-
-def test_run_subroutine_2():
-    interface = MockHostInterface()
-    processor = HostProcessor(interface, HostLatencies.all_zero(), asynchronous=True)
-
-    subrt_text = """
-    set R0 {my_value}
-    ret_reg R0
-    """
-    subrt = parse_text_subroutine(subrt_text)
     routine = LocalRoutine(
-        "subrt1",
-        subrt,
-        return_vars=["m"],
-        metadata=RoutineMetadata.use_none(),
+        "subrt1", subrt, return_vars=[IqoalaVector("res", 3)], metadata=metadata
+    )
+    instr = RunSubroutineOp(
+        result=IqoalaVector("res", 3), values=IqoalaTuple([]), subrt="subrt1"
     )
 
-    program = create_program(
-        instrs=[
-            AssignCValueOp("my_value", 16),
-            RunSubroutineOp(IqoalaVector(["m"]), IqoalaVector(["my_value"]), "subrt1"),
-        ],
-        subroutines={"subrt1": routine},
-    )
+    program = create_program(instrs=[instr], subroutines={"subrt1": routine})
     process = create_process(program, interface)
-
-    # Make sure interface can mimick writing subroutine results to shared memory
-    interface.shared_mem = process.prog_memory.shared_mem
-
     processor.initialize(process)
 
-    for i in range(len(program.instructions)):
-        yield_from(processor.assign_instr_index(process, i))
+    lrcall = processor.prepare_lr_call(process, program.instructions[0])
 
-    send_event = interface.send_events[0]
-    assert send_event.peer == "qnos"
-    assert isinstance(send_event.msg, Message)
-    assert isinstance(send_event.msg.content, LrCallTuple)
-    assert send_event.msg.content.routine_name == "subrt1"
-    assert interface.recv_events[0] == InterfaceEvent("qnos", MOCK_MESSAGE)
+    # Host processor should have allocated shared memory space.
+    assert lrcall.routine_name == "subrt1"
+    assert len(process.shared_mem.raw_arrays.raw_memory[lrcall.result_addr]) == 3
 
-    # Hack to find out which addr was allocated to write results to.
-    result_addr = process.prog_memory.shared_mem._lr_out_addrs[0]
-    assert process.prog_memory.shared_mem.read_lr_out(result_addr, 1) == [
-        MOCK_QNOS_RET_VALUE
-    ]
 
-    # Check if result was correctly written to host variable "m".
-    assert process.prog_memory.host_mem.read("m") == MOCK_QNOS_RET_VALUE
+def test_post_lr_call():
+    interface = MockHostInterface()
+    processor = HostProcessor(interface, HostLatencies.all_zero())
+
+    subrt = Subroutine()
+    metadata = RoutineMetadata.use_none()
+    routine = LocalRoutine(
+        "subrt1", subrt, return_vars=[IqoalaVector("res", 3)], metadata=metadata
+    )
+    instr = RunSubroutineOp(
+        result=IqoalaVector("res", 3), values=IqoalaTuple([]), subrt="subrt1"
+    )
+
+    program = create_program(instrs=[instr], subroutines={"subrt1": routine})
+    process = create_process(program, interface)
+    processor.initialize(process)
+
+    lrcall = processor.prepare_lr_call(process, program.instructions[0])
+    # Mock LR execution by writing results to shared memory.
+    process.shared_mem.write_lr_out(lrcall.result_addr, [1, 2, 3])
+
+    processor.post_lr_call(process, program.instructions[0], lrcall)
+
+    # Host memory should contain the results.
+    assert process.host_mem.read_vec("res") == [1, 2, 3]
 
 
 def create_simple_request(
@@ -422,43 +427,126 @@ def create_simple_request(
     )
 
 
-def test_run_request():
+def test_prepare_rr_call():
     interface = MockHostInterface()
-    processor = HostProcessor(interface, HostLatencies.all_zero(), asynchronous=True)
+    processor = HostProcessor(interface, HostLatencies.all_zero())
 
     request = create_simple_request(
-        remote_id=2,
-        num_pairs=5,
+        remote_id=0,
+        num_pairs=10,
         virt_ids=RequestVirtIdMapping.from_str("increment 0"),
-        typ=EprType.CREATE_KEEP,
+        typ=EprType.MEASURE_DIRECTLY,
         role=EprRole.CREATE,
     )
-    routine = RequestRoutine("req", request, [], CallbackType.WAIT_ALL, None)
+    routine = RequestRoutine(
+        "req", request, [IqoalaVector("outcomes", 10)], CallbackType.WAIT_ALL, None
+    )
+    instr = RunRequestOp(
+        result=IqoalaVector("outcomes", 10), values=IqoalaTuple([]), routine="req"
+    )
+
+    program = create_program(instrs=[instr], requests={"req": routine})
+    process = create_process(program, interface)
+    processor.initialize(process)
+
+    rrcall = processor.prepare_rr_call(process, program.instructions[0])
+
+    # Host processor should have allocated shared memory space.
+    assert rrcall.routine_name == "req"
+    assert len(process.shared_mem.raw_arrays.raw_memory[rrcall.result_addr]) == 10
+
+
+def test_prepare_rr_call_with_callbacks():
+    interface = MockHostInterface()
+    processor = HostProcessor(interface, HostLatencies.all_zero())
+
+    subrt = Subroutine()
+    metadata = RoutineMetadata.use_none()
+    local_routine = LocalRoutine(
+        "subrt1", subrt, return_vars=[IqoalaVector("res", 3)], metadata=metadata
+    )
+
+    request = create_simple_request(
+        remote_id=0,
+        num_pairs=10,
+        virt_ids=RequestVirtIdMapping.from_str("increment 0"),
+        typ=EprType.MEASURE_DIRECTLY,
+        role=EprRole.CREATE,
+    )
+    routine = RequestRoutine("req", request, [], CallbackType.SEQUENTIAL, "subrt1")
+
+    # We expect 10 (num_pairs) * 3 (per callback) = 30 results
+    instr = RunRequestOp(
+        result=IqoalaVector("outcomes", 30), values=IqoalaTuple([]), routine="req"
+    )
 
     program = create_program(
-        instrs=[RunRequestOp(None, IqoalaVector([]), "req")],
-        requests={"req": routine},
+        instrs=[instr], subroutines={"subrt1": local_routine}, requests={"req": routine}
     )
     process = create_process(program, interface)
     processor.initialize(process)
 
-    for i in range(len(program.instructions)):
-        yield_from(processor.assign_instr_index(process, i))
+    rrcall = processor.prepare_rr_call(process, program.instructions[0])
 
-    # Host processor should have communicated with the netstack processor.
-    send_event = interface.send_events[0]
-    assert send_event.peer == "netstack"
-    assert isinstance(send_event.msg, Message)
-    assert isinstance(send_event.msg.content, RrCallTuple)
-    assert send_event.msg.content.routine_name == "req"
-    assert interface.recv_events[0] == InterfaceEvent("netstack", MOCK_MESSAGE)
+    # Host processor should have allocated shared memory space.
+    assert rrcall.routine_name == "req"
+    raw_memory = process.shared_mem.raw_arrays.raw_memory
+    # 0 entries for the RR itself
+    assert len(raw_memory[rrcall.result_addr]) == 0
+    # 3 entries for each of the callbacks
+    assert all(len(raw_memory[addr]) == 3 for addr in rrcall.cb_output_addrs)
+
+
+def test_post_rr_call_with_callbacks():
+    interface = MockHostInterface()
+    processor = HostProcessor(interface, HostLatencies.all_zero())
+
+    subrt = Subroutine()
+    metadata = RoutineMetadata.use_none()
+    local_routine = LocalRoutine(
+        "subrt1", subrt, return_vars=[IqoalaVector("res", 3)], metadata=metadata
+    )
+
+    request = create_simple_request(
+        remote_id=0,
+        num_pairs=10,
+        virt_ids=RequestVirtIdMapping.from_str("increment 0"),
+        typ=EprType.MEASURE_DIRECTLY,
+        role=EprRole.CREATE,
+    )
+    routine = RequestRoutine("req", request, [], CallbackType.SEQUENTIAL, "subrt1")
+
+    # We expect 10 (num_pairs) * 3 (per callback) = 30 results
+    instr = RunRequestOp(
+        result=IqoalaVector("outcomes", 30), values=IqoalaTuple([]), routine="req"
+    )
+
+    program = create_program(
+        instrs=[instr], subroutines={"subrt1": local_routine}, requests={"req": routine}
+    )
+    process = create_process(program, interface)
+    processor.initialize(process)
+
+    rrcall = processor.prepare_rr_call(process, program.instructions[0])
+    # Mock RR execution by writing results to shared memory.
+    for i in range(10):
+        data = [3 * i, 3 * i + 1, 3 * i + 2]
+        process.shared_mem.write_lr_out(rrcall.cb_output_addrs[i], data)
+
+    processor.post_rr_call(process, program.instructions[0], rrcall)
+
+    # Host memory should contain the results.
+    assert process.host_mem.read_vec("outcomes") == [i for i in range(30)]
 
 
 def test_return_result():
     interface = MockHostInterface()
     processor = HostProcessor(interface, HostLatencies.all_zero())
     program = create_program(
-        instrs=[AssignCValueOp("result", 2), ReturnResultOp("result")]
+        instrs=[
+            AssignCValueOp(IqoalaSingleton("result"), 2),
+            ReturnResultOp(IqoalaSingleton("result")),
+        ]
     )
     process = create_process(program, interface)
     processor.initialize(process)
@@ -470,6 +558,41 @@ def test_return_result():
     assert process.result.values == {"result": 2}
 
 
+def test_vector_element():
+    interface = MockHostInterface()
+    processor = HostProcessor(interface, HostLatencies.all_zero())
+
+    subrt = Subroutine()
+    metadata = RoutineMetadata.use_none()
+    routine = LocalRoutine(
+        "subrt1", subrt, return_vars=[IqoalaVector("res", 3)], metadata=metadata
+    )
+    instr = RunSubroutineOp(
+        result=IqoalaVector("res", 3), values=IqoalaTuple([]), subrt="subrt1"
+    )
+    instr2 = AddCValueOp(
+        IqoalaSingleton("result"),
+        IqoalaVectorElement("res", 0),
+        IqoalaVectorElement("res", 1),
+    )
+
+    program = create_program(instrs=[instr, instr2], subroutines={"subrt1": routine})
+    process = create_process(program, interface)
+    processor.initialize(process)
+
+    lrcall = processor.prepare_lr_call(process, program.instructions[0])
+    # Mock LR execution by writing results to shared memory.
+    process.shared_mem.write_lr_out(lrcall.result_addr, [1, 2, 3])
+
+    processor.post_lr_call(process, program.instructions[0], lrcall)
+
+    yield_from(processor.assign_instr_index(process, 1))
+
+    # Host memory should contain the results.
+    assert process.host_mem.read_vec("res") == [1, 2, 3]
+    assert process.prog_memory.host_mem.read("result") == 3
+
+
 if __name__ == "__main__":
     test_initialize()
     test_assign_cvalue()
@@ -481,7 +604,10 @@ if __name__ == "__main__":
     test_add_cvalue_with_latencies()
     test_multiply_const()
     test_bit_cond_mult()
-    test_run_subroutine()
-    test_run_subroutine_2()
-    test_run_request()
+    test_prepare_lr_call()
+    test_post_lr_call()
+    test_prepare_rr_call()
+    test_prepare_rr_call_with_callbacks()
+    test_post_rr_call_with_callbacks()
     test_return_result()
+    test_vector_element()
