@@ -8,8 +8,17 @@ from netqasm.lang.subroutine import Subroutine
 from qoala.lang.hostlang import (
     AssignCValueOp,
     BasicBlockType,
+    BusyOp,
+    IqoalaSingleton,
+    IqoalaTuple,
     IqoalaVector,
+    IqoalaVectorElement,
+    MultiplyConstantCValueOp,
+    ReceiveCMsgOp,
+    ReturnResultOp,
+    RunRequestOp,
     RunSubroutineOp,
+    SendCMsgOp,
 )
 from qoala.lang.parse import (
     HostCodeParser,
@@ -104,7 +113,7 @@ x = assign_cval() : 1
     instructions = IqoalaInstrParser(text).parse()
 
     assert len(instructions) == 1
-    assert instructions[0] == AssignCValueOp(result="x", value=1)
+    assert instructions[0] == AssignCValueOp(result=IqoalaSingleton("x"), value=1)
 
 
 def test_parse_2_instr():
@@ -116,8 +125,19 @@ y = assign_cval() : 17
     instructions = IqoalaInstrParser(text).parse()
 
     assert len(instructions) == 2
-    assert instructions[0] == AssignCValueOp(result="x", value=1)
-    assert instructions[1] == AssignCValueOp(result="y", value=17)
+    assert instructions[0] == AssignCValueOp(result=IqoalaSingleton("x"), value=1)
+    assert instructions[1] == AssignCValueOp(result=IqoalaSingleton("y"), value=17)
+
+
+def test_parse_busy():
+    text = """
+busy() : 250
+    """
+
+    instructions = IqoalaInstrParser(text).parse()
+
+    assert len(instructions) == 1
+    assert instructions[0] == BusyOp(value=250)
 
 
 def test_parse_faulty_instr():
@@ -129,62 +149,108 @@ x = assign_cval
         IqoalaInstrParser(text).parse()
 
 
-def test_parse_vec():
+def test_parse_tuple():
     text = """
-run_subroutine(vec<x>) : subrt1
+run_subroutine(tuple<x>) : subrt1
     """
 
     instructions = IqoalaInstrParser(text).parse()
     assert len(instructions) == 1
     assert instructions[0] == RunSubroutineOp(
-        result=None, values=IqoalaVector(["x"]), subrt="subrt1"
+        result=None, values=IqoalaTuple(["x"]), subrt="subrt1"
     )
 
 
-def test_parse_vec_2_elements():
+def test_parse_tuple_2_elements():
     text = """
-run_subroutine(vec<x; y>) : subrt1
+run_subroutine(tuple<x; y>) : subrt1
     """
 
     instructions = IqoalaInstrParser(text).parse()
     assert len(instructions) == 1
     assert instructions[0] == RunSubroutineOp(
-        result=None, values=IqoalaVector(["x", "y"]), subrt="subrt1"
+        result=None, values=IqoalaTuple(["x", "y"]), subrt="subrt1"
     )
 
 
-def test_parse_vec_2_elements_and_return():
+def test_parse_tuple_2_elements_and_return():
     text = """
-vec<m> = run_subroutine(vec<x; y>) : subrt1
+tuple<m> = run_subroutine(tuple<x; y>) : subrt1
     """
 
     instructions = IqoalaInstrParser(text).parse()
     assert len(instructions) == 1
     assert instructions[0] == RunSubroutineOp(
-        result=IqoalaVector(["m"]), values=IqoalaVector(["x", "y"]), subrt="subrt1"
+        result=IqoalaTuple(["m"]), values=IqoalaTuple(["x", "y"]), subrt="subrt1"
     )
 
 
-def test_parse_vec_2_elements_and_return_2_elements():
+def test_parse_tuple_2_elements_and_return_2_elements():
     text = """
-vec<m1; m2> = run_subroutine(vec<x; y>) : subrt1
+tuple<m1; m2> = run_subroutine(tuple<x; y>) : subrt1
     """
 
     instructions = IqoalaInstrParser(text).parse()
     assert len(instructions) == 1
     assert instructions[0] == RunSubroutineOp(
-        result=IqoalaVector(["m1", "m2"]),
-        values=IqoalaVector(["x", "y"]),
+        result=IqoalaTuple(["m1", "m2"]),
+        values=IqoalaTuple(["x", "y"]),
         subrt="subrt1",
     )
 
 
+def test_parse_vector():
+    text = """
+my_vec<3> = run_subroutine() : subrt1
+    """
+
+    instructions = IqoalaInstrParser(text).parse()
+    assert len(instructions) == 1
+    assert instructions[0] == RunSubroutineOp(
+        result=IqoalaVector("my_vec", 3), values=IqoalaTuple([]), subrt="subrt1"
+    )
+
+
+def test_parse_vector_2():
+    text = """
+my_vec<3> = run_request() : req1
+    """
+
+    instructions = IqoalaInstrParser(text).parse()
+    assert len(instructions) == 1
+    assert instructions[0] == RunRequestOp(
+        result=IqoalaVector("my_vec", 3), values=IqoalaTuple([]), routine="req1"
+    )
+
+
+def test_parse_vector_with_var():
+    text = """
+my_vec<N> = run_request() : req1
+    """
+
+    instructions = IqoalaInstrParser(text).parse()
+    assert len(instructions) == 1
+    assert instructions[0] == RunRequestOp(
+        result=IqoalaVector("my_vec", "N"), values=IqoalaTuple([]), routine="req1"
+    )
+
+
 def test_parse_block_header():
     text = "^b0 {type = CL}:"
 
-    name, typ = HostCodeParser("")._parse_block_header(text)
+    name, typ, duration = HostCodeParser("")._parse_block_header(text)
     assert name == "b0"
     assert typ == BasicBlockType.CL
+    assert duration is None
+
+
+def test_parse_block_header_with_deadlines():
+    text = "^b0 {type = CL, deadlines = [b1: 1000]}:"
+
+    name, typ, deadline = HostCodeParser("")._parse_block_header(text)
+    assert name == "b0"
+    assert typ == BasicBlockType.CL
+    assert deadline == {"b1": 1000}
 
 
 def test_parse_block():
@@ -198,8 +264,10 @@ def test_parse_block():
     assert block.name == "b0"
     assert block.typ == BasicBlockType.CL
     assert len(block.instructions) == 2
-    assert block.instructions[0] == AssignCValueOp(result="x", value=1)
-    assert block.instructions[1] == AssignCValueOp(result="y", value=17)
+    assert block.instructions[0] == AssignCValueOp(result=IqoalaSingleton("x"), value=1)
+    assert block.instructions[1] == AssignCValueOp(
+        result=IqoalaSingleton("y"), value=17
+    )
 
 
 def test_get_block_texts():
@@ -209,7 +277,7 @@ def test_get_block_texts():
     y = assign_cval() : 17
 
 ^b1 {type = QL}:
-    run_subroutine(vec<x>) : subrt1
+    run_subroutine(tuple<x>) : subrt1
     """
 
     block_texts = HostCodeParser(text).get_block_texts()
@@ -222,8 +290,8 @@ def test_parse_multiple_blocks():
     x = assign_cval() : 1
     y = assign_cval() : 17
 
-^b1 {type = QL}:
-    run_subroutine(vec<x>) : subrt1
+^b1 {type = QL, deadlines = [b0: 2500]}:
+    run_subroutine(tuple<x>) : subrt1
     """
 
     blocks = HostCodeParser(text).parse()
@@ -232,14 +300,19 @@ def test_parse_multiple_blocks():
     assert blocks[0].name == "b0"
     assert blocks[0].typ == BasicBlockType.CL
     assert len(blocks[0].instructions) == 2
-    assert blocks[0].instructions[0] == AssignCValueOp(result="x", value=1)
-    assert blocks[0].instructions[1] == AssignCValueOp(result="y", value=17)
+    assert blocks[0].instructions[0] == AssignCValueOp(
+        result=IqoalaSingleton("x"), value=1
+    )
+    assert blocks[0].instructions[1] == AssignCValueOp(
+        result=IqoalaSingleton("y"), value=17
+    )
 
     assert blocks[1].name == "b1"
     assert blocks[1].typ == BasicBlockType.QL
+    assert blocks[1].deadlines == {"b0": 2500}
     assert len(blocks[1].instructions) == 1
     assert blocks[1].instructions[0] == RunSubroutineOp(
-        result=None, values=IqoalaVector(["x"]), subrt="subrt1"
+        result=None, values=IqoalaTuple(["x"]), subrt="subrt1"
     )
 
 
@@ -267,7 +340,7 @@ SUBROUTINE subrt1
     expected_args = ["my_value"]
     assert subrt == LocalRoutine(
         name="subrt1",
-        subrt=Subroutine(instructions=expected_instrs, arguments=expected_args),
+        subroutine=Subroutine(instructions=expected_instrs, arguments=expected_args),
         metadata=RoutineMetadata.use_none(),
         return_vars=["m"],
     )
@@ -303,12 +376,48 @@ SUBROUTINE my_subroutine
     expected_args = ["param1", "param2"]
     assert subrt == LocalRoutine(
         name="my_subroutine",
-        subrt=Subroutine(instructions=expected_instrs, arguments=expected_args),
+        subroutine=Subroutine(instructions=expected_instrs, arguments=expected_args),
         metadata=RoutineMetadata.free_all([0, 1]),
         return_vars=["result1", "result2"],
     )
 
 
+def test_parse_subrt_3():
+    text = """
+SUBROUTINE my_subroutine
+    params: param1, param2
+    returns: outcomes<10>
+    uses: 0, 1
+    keeps: 
+    request: 
+  NETQASM_START
+    set R0 {param1}
+    set R1 {param2}
+    meas Q0 M5
+    meas Q1 M6
+  NETQASM_END
+    """
+
+    parsed = LocalRoutineParser(text).parse()
+    assert len(parsed) == 1
+    assert "my_subroutine" in parsed
+    subrt = parsed["my_subroutine"]
+
+    expected_instrs = [
+        SetInstruction(reg=Register.from_str("R0"), imm=Template("param1")),
+        SetInstruction(reg=Register.from_str("R1"), imm=Template("param2")),
+        MeasInstruction(reg0=Register.from_str("Q0"), reg1=Register.from_str("M5")),
+        MeasInstruction(reg0=Register.from_str("Q1"), reg1=Register.from_str("M6")),
+    ]
+    expected_args = ["param1", "param2"]
+    assert subrt == LocalRoutine(
+        name="my_subroutine",
+        subroutine=Subroutine(instructions=expected_instrs, arguments=expected_args),
+        metadata=RoutineMetadata.free_all([0, 1]),
+        return_vars=[IqoalaVector("outcomes", 10)],
+    )
+
+
 def test_parse_multiple_subrt():
     text = """
 SUBROUTINE subrt1
@@ -352,13 +461,17 @@ SUBROUTINE subrt2
 
     assert subrt1 == LocalRoutine(
         name="subrt1",
-        subrt=Subroutine(instructions=expected_instrs_1, arguments=expected_args_1),
+        subroutine=Subroutine(
+            instructions=expected_instrs_1, arguments=expected_args_1
+        ),
         metadata=RoutineMetadata.free_all([0]),
         return_vars=["m"],
     )
     assert subrt2 == LocalRoutine(
         name="subrt2",
-        subrt=Subroutine(instructions=expected_instrs_2, arguments=expected_args_2),
+        subroutine=Subroutine(
+            instructions=expected_instrs_2, arguments=expected_args_2
+        ),
         metadata=RoutineMetadata.use_none(),
         return_vars=[],
     )
@@ -424,7 +537,7 @@ def test_parse_request_2():
 REQUEST req1
   callback_type: sequential
   callback: subrt1
-  return_vars: 
+  return_vars: outcomes<3>
   remote_id: 1
   epr_socket_id: 0
   num_pairs: 3
@@ -442,7 +555,7 @@ REQUEST req1
 
     assert routine == RequestRoutine(
         name="req1",
-        return_vars=[],
+        return_vars=[IqoalaVector("outcomes", 3)],
         callback_type=CallbackType.SEQUENTIAL,
         callback="subrt1",
         request=QoalaRequest(
@@ -499,6 +612,46 @@ REQUEST req1
     )
 
 
+def test_parse_request_with_vector_template():
+    text = """
+REQUEST req1
+  callback_type: wait_all
+  callback: 
+  return_vars: outcomes<{N}>
+  remote_id: 0
+  epr_socket_id: 0
+  num_pairs: 3
+  virt_ids: custom 1, 2, 3
+  timeout: 1000
+  fidelity: 0.65
+  typ: measure_directly
+  role: receive
+    """
+
+    parsed = RequestRoutineParser(text).parse()
+    assert len(parsed) == 1
+    assert "req1" in parsed
+    routine = parsed["req1"]
+
+    assert routine == RequestRoutine(
+        name="req1",
+        return_vars=[IqoalaVector("outcomes", Template("N"))],
+        callback_type=CallbackType.WAIT_ALL,
+        callback=None,
+        request=QoalaRequest(
+            name="req1",
+            remote_id=0,
+            epr_socket_id=0,
+            num_pairs=3,
+            virt_ids=RequestVirtIdMapping.from_str("custom 1, 2, 3"),
+            timeout=1000,
+            fidelity=0.65,
+            typ=EprType.MEASURE_DIRECTLY,
+            role=EprRole.RECEIVE,
+        ),
+    )
+
+
 def test_parse_multiple_request():
     text = """
 REQUEST req1
@@ -623,10 +776,10 @@ META_END
     my_value = add_cval_c(new_value, new_value)
 
 ^b1 {type = QL}:
-    vec<m> = run_subroutine(vec<my_value>) : subrt1
+    tuple<m> = run_subroutine(tuple<my_value>) : subrt1
 
 ^b2 {type = QC}:
-    vec<m> = run_request(vec<my_value>) : req1
+    tuple<m> = run_request(tuple<my_value>) : req1
 
 ^b3 {type = CL}:
     return_result(m)
@@ -695,7 +848,7 @@ META_END
     my_value = add_cval_c(new_value, new_value)
 
 ^b1 {type = QL}:
-    vec<m> = run_subroutine(vec<my_value>) : subrt1
+    tuple<m> = run_subroutine(tuple<my_value>) : subrt1
 
 ^b1 {type = CL}:
     return_result(m)
@@ -783,7 +936,7 @@ META_END
 ^b0 {type = CL}:
     my_value = assign_cval() : 1
 ^b1 {type = QL}:
-    vec<m> = run_subroutine(vec<my_value>) : non_existing_subrt
+    tuple<m> = run_subroutine(tuple<my_value>) : non_existing_subrt
 ^b2 {type = CL}:
     return_result(m)
     """
@@ -823,7 +976,7 @@ META_END
 ^b0 {type = CL}:
     my_value = assign_cval() : 1
 ^b1 {type = QL}:
-    vec<m> = run_request(vec<my_value>) : non_existing_req
+    tuple<m> = run_request(tuple<my_value>) : non_existing_req
 ^b2 {type = CL}:
     return_result(m)
     """
@@ -959,7 +1112,7 @@ META_END
     ^b0 {type = CL}:
         remote_id = assign_cval() : {client_id}
     ^b1 {type = QC}:
-        run_request(vec<>) : req1
+        run_request() : req1
         """
 
     req_text = """
@@ -1002,7 +1155,7 @@ META_END
     new_value = assign_cval() : 3
     my_value = add_cval_c(new_value, new_value)
 ^b1 {type = QL}:
-    vec<m> = run_subroutine(vec<my_value>) : subrt1
+    tuple<m> = run_subroutine(tuple<my_value>) : subrt1
 ^b2 {type = CL}:
     return_result(m)
 
@@ -1061,6 +1214,53 @@ def test_parse_file_2():
     assert parsed_program.local_routines["create_epr_1"].request_name == "req1"
 
 
+def test_vector_indexing():
+    text = """
+    a<3> = run_request() : req1    
+    send_cmsg(a[0], b)
+    b = recv_cmsg(a[0])
+    b = mult_const(a[0]) : 1
+    """
+
+    instructions = IqoalaInstrParser(text).parse()
+
+    assert len(instructions) == 4
+    csocket = IqoalaVectorElement("a", 0)
+    remote_node = IqoalaSingleton("b")
+    assert instructions[0] == RunRequestOp(
+        result=IqoalaVector("a", 3), values=IqoalaTuple([]), routine="req1"
+    )
+    assert instructions[1] == SendCMsgOp(csocket, remote_node)
+    assert instructions[2] == ReceiveCMsgOp(csocket, remote_node)
+    assert instructions[3] == MultiplyConstantCValueOp(remote_node, csocket, 1)
+
+
+def test_vector_indexing_error():
+    text = """
+    send_cmsg(a[0], b)
+    b = recv_cmsg(a[0])
+    b = mult_const(a[0]) : 1
+    """
+
+    with pytest.raises(QoalaParseError):
+        IqoalaInstrParser(text).parse()
+
+
+def test_arg_vector_passing():
+    text = """
+    a<3> = run_request() : req1    
+    return_result(a)
+    """
+
+    instructions = IqoalaInstrParser(text).parse()
+    assert len(instructions) == 2
+    vec = IqoalaVector("a", 3)
+    assert instructions[0] == RunRequestOp(
+        result=vec, values=IqoalaTuple([]), routine="req1"
+    )
+    assert instructions[1] == ReturnResultOp(vec)
+
+
 if __name__ == "__main__":
     test_parse_incomplete_meta()
     test_parse_meta_no_end()
@@ -1068,22 +1268,29 @@ if __name__ == "__main__":
     test_parse_meta_multiple_remotes()
     test_parse_1_instr()
     test_parse_2_instr()
+    test_parse_busy()
     test_parse_faulty_instr()
-    test_parse_vec()
-    test_parse_vec_2_elements()
-    test_parse_vec_2_elements_and_return()
-    test_parse_vec_2_elements_and_return_2_elements()
+    test_parse_tuple()
+    test_parse_tuple_2_elements()
+    test_parse_tuple_2_elements_and_return()
+    test_parse_tuple_2_elements_and_return_2_elements()
+    test_parse_vector()
+    test_parse_vector_2()
+    test_parse_vector_with_var()
     test_parse_block_header()
+    test_parse_block_header_with_deadlines()
     test_parse_block()
     test_get_block_texts()
     test_parse_multiple_blocks()
     test_parse_subrt()
     test_parse_subrt_2()
+    test_parse_subrt_3()
     test_parse_multiple_subrt()
     test_parse_invalid_subrt()
     test_parse_request()
     test_parse_request_2()
     test_parse_request_with_template()
+    test_parse_request_with_vector_template()
     test_parse_multiple_request()
     test_parse_invalid_request()
     test_parse_program()
@@ -1097,3 +1304,6 @@ if __name__ == "__main__":
     test_parse_program_single_text()
     test_parse_file()
     test_parse_file_2()
+    test_vector_indexing()
+    test_vector_indexing_error()
+    test_arg_vector_passing()
