@@ -15,9 +15,8 @@ from qoala.lang.hostlang import (
     BasicBlock,
     BasicBlockType,
     ClassicalIqoalaOp,
-    IqoalaVector,
+    IqoalaSingleton,
     ReceiveCMsgOp,
-    RunSubroutineOp,
     SendCMsgOp,
 )
 from qoala.lang.parse import LocalRoutineParser, QoalaParser
@@ -31,21 +30,16 @@ from qoala.lang.request import (
     RequestVirtIdMapping,
 )
 from qoala.lang.routine import RoutineMetadata
-from qoala.runtime.config import GenericQDeviceConfig
-from qoala.runtime.environment import LocalEnvironment, NetworkInfo
 from qoala.runtime.lhi import LhiLatencies, LhiLinkInfo, LhiTopology, LhiTopologyBuilder
-from qoala.runtime.lhi_to_ehi import (
-    GenericToVanillaInterface,
-    LhiConverter,
-    NativeToFlavourInterface,
-)
+from qoala.runtime.lhi_to_ehi import LhiConverter
 from qoala.runtime.memory import ProgramMemory
 from qoala.runtime.message import Message, RrCallTuple
+from qoala.runtime.ntf import GenericNtf, NtfInterface
 from qoala.runtime.program import ProgramInput, ProgramInstance, ProgramResult
-from qoala.sim.build import build_generic_qprocessor
+from qoala.sim.build import build_qprocessor_from_topology
 from qoala.sim.entdist.entdist import EntDist
 from qoala.sim.entdist.entdistcomp import EntDistComponent
-from qoala.sim.events import MSG_REQUEST_DELIVERED
+from qoala.sim.eprsocket import EprSocket
 from qoala.sim.host.csocket import ClassicalSocket
 from qoala.sim.host.hostinterface import HostInterface
 from qoala.sim.memmgr import MemoryManager
@@ -53,26 +47,27 @@ from qoala.sim.netstack import NetstackInterface
 from qoala.sim.process import QoalaProcess
 from qoala.sim.procnode import ProcNode
 from qoala.sim.qdevice import QDevice, QDeviceCommand
+from qoala.sim.scheduler import NodeScheduler
 from qoala.util.math import has_multi_state
 from qoala.util.tests import netsquid_run
 
-MOCK_MESSAGE = Message(content=42)
+MOCK_MESSAGE = Message(src_pid=0, dst_pid=0, content=42)
 MOCK_QNOS_RET_REG = "R0"
 MOCK_QNOS_RET_VALUE = 7
 
 
-@dataclass(eq=True, frozen=True)
+@dataclass(frozen=True)
 class InterfaceEvent:
     peer: str
     msg: Message
 
 
-@dataclass(eq=True, frozen=True)
+@dataclass(frozen=True)
 class FlushEvent:
     pass
 
 
-@dataclass(eq=True, frozen=True)
+@dataclass(frozen=True)
 class SignalEvent:
     pass
 
@@ -80,12 +75,10 @@ class SignalEvent:
 class MockNetstackInterface(NetstackInterface):
     def __init__(
         self,
-        local_env: LocalEnvironment,
         qdevice: QDevice,
         memmgr: MemoryManager,
     ) -> None:
         self._qdevice = qdevice
-        self._local_env = local_env
         self._memmgr = memmgr
 
     def send_qnos_msg(self, msg: Message) -> None:
@@ -102,7 +95,7 @@ class MockNetstackInterface(NetstackInterface):
         return None
 
     def receive_entdist_msg(self) -> Generator[EventExpression, None, Message]:
-        return Message(MSG_REQUEST_DELIVERED)
+        return Message(-1, -1, 0)
         yield  # to make this behave as a generator
 
     @property
@@ -134,6 +127,11 @@ class MockHostInterface(HostInterface):
     def __init__(self) -> None:
         self.send_events: List[InterfaceEvent] = []
         self.recv_events: List[InterfaceEvent] = []
+        self._program_instance_jumps: Dict[int, int] = {}  # pid => block name
+
+    @property
+    def program_instance_jumps(self) -> Dict[int, int]:
+        return self._program_instance_jumps
 
     def send_peer_msg(self, peer: str, msg: Message) -> None:
         self.send_events.append(InterfaceEvent(peer, msg))
@@ -196,7 +194,6 @@ def create_process(
         program=program,
         inputs=prog_input,
         unit_module=unit_module,
-        block_tasks=[],
     )
     mem = ProgramMemory(pid=0)
 
@@ -204,48 +201,65 @@ def create_process(
         prog_instance=instance,
         prog_memory=mem,
         csockets={
-            id: ClassicalSocket(host_interface, name)
+            id: ClassicalSocket(host_interface, name, 0, 0)
             for (id, name) in program.meta.csockets.items()
         },
-        epr_sockets=program.meta.epr_sockets,
+        epr_sockets={0: EprSocket(0, 0, 0, 0, 1.0)},
         result=ProgramResult(values={}),
     )
     return process
 
 
+class MockScheduler(NodeScheduler):
+    def __init__(self, scheduler: NodeScheduler) -> None:
+        self._cpu_scheduler = scheduler.cpu_scheduler
+        self._qpu_scheduler = scheduler.qpu_scheduler
+        self._host = scheduler.host
+        self._last_cpu_task_pid = -1
+        self._last_qpu_task_pid = -1
+        self._is_predictable = True
+        pass
+
+    def schedule_next_for(self, pid: int) -> None:
+        pass
+
+    def schedule_all(self) -> None:
+        pass
+
+
 def create_qprocessor(name: str, num_qubits: int) -> QuantumProcessor:
-    cfg = GenericQDeviceConfig.perfect_config(num_qubits=num_qubits)
-    return build_generic_qprocessor(name=f"{name}_processor", cfg=cfg)
+    topology = LhiTopologyBuilder.perfect_uniform_default_gates(num_qubits=num_qubits)
+    return build_qprocessor_from_topology(name=f"{name}_processor", topology=topology)
 
 
-def create_network_info(
-    num_qubits: int, names: List[str] = ["alice", "bob", "charlie"]
-) -> NetworkInfo:
+def create_network_ehi(num_qubits: int, names: List[str] = None) -> EhiNetworkInfo:
+    if names is None:
+        names = ["alice", "bob", "charlie"]
+
     nodes = {i: name for i, name in enumerate(names)}
-    return NetworkInfo.with_nodes(nodes)
+    return EhiNetworkInfo(nodes, links={})
 
 
 def create_procnode(
     name: str,
-    env: NetworkInfo,
+    network_ehi: EhiNetworkInfo,
     num_qubits: int,
     topology: LhiTopology,
     latencies: LhiLatencies,
-    ntf_interface: NativeToFlavourInterface,
+    ntf_interface: NtfInterface,
     procnode_cls: Type[ProcNode] = ProcNode,
     asynchronous: bool = False,
 ) -> ProcNode:
     alice_qprocessor = create_qprocessor(name, num_qubits)
 
-    node_id = env.get_node_id(name)
+    node_id = network_ehi.get_node_id(name)
     procnode = procnode_cls(
         name=name,
-        network_info=env,
         qprocessor=alice_qprocessor,
         qdevice_topology=topology,
         latencies=latencies,
         ntf_interface=ntf_interface,
-        network_ehi=EhiNetworkInfo({}),
+        network_ehi=network_ehi,
         node_id=node_id,
         asynchronous=asynchronous,
     )
@@ -292,19 +306,18 @@ def test_initialize():
     num_qubits = 3
     topology = generic_topology(num_qubits)
     latencies = LhiLatencies.all_zero()
-    ntf = GenericToVanillaInterface()
+    ntf = GenericNtf()
 
-    network_info = create_network_info(num_qubits)
-    local_env = LocalEnvironment(network_info, network_info.get_node_id("alice"))
+    network_ehi = create_network_ehi(num_qubits)
     procnode = create_procnode(
-        "alice", network_info, num_qubits, topology, latencies, ntf
+        "alice", network_ehi, num_qubits, topology, latencies, ntf
     )
     procnode.qdevice = MockQDevice(topology)
 
     procnode.host.interface = MockHostInterface()
 
     procnode.netstack.interface = MockNetstackInterface(
-        local_env, procnode.qdevice, procnode.memmgr
+        procnode.qdevice, procnode.memmgr
     )
 
     host_processor = procnode.host.processor
@@ -314,7 +327,7 @@ def test_initialize():
     ehi = LhiConverter.to_ehi(topology, ntf, latencies)
     unit_module = UnitModule.from_full_ehi(ehi)
 
-    instrs = [AssignCValueOp("x", 3)]
+    instrs = [AssignCValueOp(IqoalaSingleton("x"), 3)]
     subrt1 = simple_subroutine(
         "subrt1",
         """
@@ -373,69 +386,6 @@ def test_initialize():
     assert procnode.memmgr.phys_id_for(pid=process.pid, virt_id=0) == 0
 
 
-def test_2():
-    ns.sim_reset()
-
-    num_qubits = 3
-    topology = generic_topology(num_qubits)
-    latencies = LhiLatencies.all_zero()
-    ntf = GenericToVanillaInterface()
-
-    network_info = create_network_info(num_qubits)
-    procnode = create_procnode(
-        "alice", network_info, num_qubits, topology, latencies, ntf, asynchronous=True
-    )
-    procnode.qdevice = MockQDevice(topology)
-
-    host_processor = procnode.host.processor
-    qnos_processor = procnode.qnos.processor
-
-    ehi = LhiConverter.to_ehi(topology, ntf, latencies)
-    unit_module = UnitModule.from_full_ehi(ehi)
-
-    instrs = [RunSubroutineOp(IqoalaVector(["result"]), IqoalaVector([]), "subrt1")]
-    subroutines = parse_iqoala_subroutines(
-        """
-SUBROUTINE subrt1
-    params: 
-    returns: result
-    uses: 
-    keeps: 
-    request:
-  NETQASM_START
-    set R5 42
-    store R5 @output[0]
-  NETQASM_END
-    """
-    )
-
-    program = create_program(instrs=instrs, subroutines=subroutines)
-    process = create_process(
-        pid=0,
-        program=program,
-        unit_module=unit_module,
-        host_interface=procnode.host._interface,
-        inputs={"x": 1, "theta": 3.14, "name": "alice"},
-    )
-    procnode.add_process(process)
-
-    host_processor.initialize(process)
-
-    def host_run() -> Generator[EventExpression, None, None]:
-        yield from host_processor.assign_instr_index(process, instr_idx=0)
-
-    def qnos_run() -> Generator[EventExpression, None, None]:
-        yield from qnos_processor.await_local_routine_call(process)
-
-    procnode.host.run = host_run
-    procnode.qnos.run = qnos_run
-    procnode.start()
-    ns.sim_run()
-
-    assert process.host_mem.read("result") == 42
-    assert process.qnos_mem.get_reg_value("R5") == 42
-
-
 def test_classical_comm():
     ns.sim_reset()
 
@@ -443,9 +393,9 @@ def test_classical_comm():
 
     topology = generic_topology(num_qubits)
     latencies = LhiLatencies.all_zero()
-    ntf = GenericToVanillaInterface()
+    ntf = GenericNtf()
 
-    network_info = create_network_info(num_qubits)
+    network_ehi = create_network_ehi(num_qubits)
 
     class TestProcNode(ProcNode):
         def run(self) -> Generator[EventExpression, None, None]:
@@ -454,7 +404,7 @@ def test_classical_comm():
 
     alice_procnode = create_procnode(
         "alice",
-        network_info,
+        network_ehi,
         num_qubits,
         topology,
         latencies,
@@ -463,7 +413,7 @@ def test_classical_comm():
     )
     bob_procnode = create_procnode(
         "bob",
-        network_info,
+        network_ehi,
         num_qubits,
         topology,
         latencies,
@@ -477,7 +427,9 @@ def test_classical_comm():
     ehi = LhiConverter.to_ehi(topology, ntf, latencies)
     unit_module = UnitModule.from_full_ehi(ehi)
 
-    alice_instrs = [SendCMsgOp("csocket_id", "message")]
+    alice_instrs = [
+        SendCMsgOp(IqoalaSingleton("csocket_id"), IqoalaSingleton("message"))
+    ]
     alice_meta = ProgramMeta(
         name="alice",
         parameters=["csocket_id", "message"],
@@ -495,7 +447,9 @@ def test_classical_comm():
     alice_procnode.add_process(alice_process)
     alice_host_processor.initialize(alice_process)
 
-    bob_instrs = [ReceiveCMsgOp("csocket_id", "result")]
+    bob_instrs = [
+        ReceiveCMsgOp(IqoalaSingleton("csocket_id"), IqoalaSingleton("result"))
+    ]
     bob_meta = ProgramMeta(
         name="bob", parameters=["csocket_id"], csockets={0: "alice"}, epr_sockets={}
     )
@@ -512,11 +466,17 @@ def test_classical_comm():
 
     alice_procnode.connect_to(bob_procnode)
 
+    # To prevent scheduler from running
+    alice_procnode.scheduler = MockScheduler(alice_procnode.scheduler)
+    bob_procnode.scheduler = MockScheduler(bob_procnode.scheduler)
+
     # First start Bob, since Alice won't yield on anything (she only does a Send
     # instruction) and therefore calling 'start()' on alice completes her whole
     # protocol while Bob's interface has not even been started.
+
     bob_procnode.start()
     alice_procnode.start()
+
     ns.sim_run()
 
     assert bob_process.host_mem.read("result") == 1337
@@ -529,9 +489,9 @@ def test_classical_comm_three_nodes():
 
     topology = generic_topology(num_qubits)
     latencies = LhiLatencies.all_zero()
-    ntf = GenericToVanillaInterface()
+    ntf = GenericNtf()
 
-    network_info = create_network_info(num_qubits)
+    network_ehi = create_network_ehi(num_qubits)
 
     class SenderProcNode(ProcNode):
         def run(self) -> Generator[EventExpression, None, None]:
@@ -546,7 +506,7 @@ def test_classical_comm_three_nodes():
 
     alice_procnode = create_procnode(
         "alice",
-        network_info,
+        network_ehi,
         num_qubits,
         topology,
         latencies,
@@ -555,7 +515,7 @@ def test_classical_comm_three_nodes():
     )
     bob_procnode = create_procnode(
         "bob",
-        network_info,
+        network_ehi,
         num_qubits,
         topology,
         latencies,
@@ -564,7 +524,7 @@ def test_classical_comm_three_nodes():
     )
     charlie_procnode = create_procnode(
         "charlie",
-        network_info,
+        network_ehi,
         num_qubits,
         topology,
         latencies,
@@ -579,7 +539,9 @@ def test_classical_comm_three_nodes():
     ehi = LhiConverter.to_ehi(topology, ntf, latencies)
     unit_module = UnitModule.from_full_ehi(ehi)
 
-    alice_instrs = [SendCMsgOp("csocket_id", "message")]
+    alice_instrs = [
+        SendCMsgOp(IqoalaSingleton("csocket_id"), IqoalaSingleton("message"))
+    ]
     alice_meta = ProgramMeta(
         name="alice",
         parameters=["csocket_id", "message"],
@@ -597,7 +559,7 @@ def test_classical_comm_three_nodes():
     alice_procnode.add_process(alice_process)
     alice_host_processor.initialize(alice_process)
 
-    bob_instrs = [SendCMsgOp("csocket_id", "message")]
+    bob_instrs = [SendCMsgOp(IqoalaSingleton("csocket_id"), IqoalaSingleton("message"))]
     bob_meta = ProgramMeta(
         name="bob",
         parameters=["csocket_id", "message"],
@@ -616,8 +578,10 @@ def test_classical_comm_three_nodes():
     bob_host_processor.initialize(bob_process)
 
     charlie_instrs = [
-        ReceiveCMsgOp("csocket_id_alice", "result_alice"),
-        ReceiveCMsgOp("csocket_id_bob", "result_bob"),
+        ReceiveCMsgOp(
+            IqoalaSingleton("csocket_id_alice"), IqoalaSingleton("result_alice")
+        ),
+        ReceiveCMsgOp(IqoalaSingleton("csocket_id_bob"), IqoalaSingleton("result_bob")),
     ]
     charlie_meta = ProgramMeta(
         name="bob",
@@ -639,6 +603,11 @@ def test_classical_comm_three_nodes():
     alice_procnode.connect_to(charlie_procnode)
     bob_procnode.connect_to(charlie_procnode)
 
+    # To prevent scheduler from running
+    alice_procnode.scheduler = MockScheduler(alice_procnode.scheduler)
+    bob_procnode.scheduler = MockScheduler(bob_procnode.scheduler)
+    charlie_procnode.scheduler = MockScheduler(charlie_procnode.scheduler)
+
     # First start Charlie, since Alice and Bob don't yield on anything.
     charlie_procnode.start()
     alice_procnode.start()
@@ -656,11 +625,11 @@ def test_epr():
 
     topology = generic_topology(num_qubits)
     latencies = LhiLatencies.all_zero()
-    ntf = GenericToVanillaInterface()
+    ntf = GenericNtf()
 
-    network_info = create_network_info(num_qubits)
-    alice_id = network_info.get_node_id("alice")
-    bob_id = network_info.get_node_id("bob")
+    network_ehi = create_network_ehi(num_qubits)
+    alice_id = network_ehi.get_node_id("alice")
+    bob_id = network_ehi.get_node_id("bob")
 
     class TestProcNode(ProcNode):
         def run(self) -> Generator[EventExpression, None, None]:
@@ -670,7 +639,7 @@ def test_epr():
 
     alice_procnode = create_procnode(
         "alice",
-        network_info,
+        network_ehi,
         num_qubits,
         topology,
         latencies,
@@ -679,7 +648,7 @@ def test_epr():
     )
     bob_procnode = create_procnode(
         "bob",
-        network_info,
+        network_ehi,
         num_qubits,
         topology,
         latencies,
@@ -771,15 +740,19 @@ def test_epr():
     alice_procnode.connect_to(bob_procnode)
 
     nodes = [alice_procnode.node, bob_procnode.node]
-    entdistcomp = EntDistComponent(network_info)
+    entdistcomp = EntDistComponent(network_ehi)
     alice_procnode.node.entdist_out_port.connect(entdistcomp.node_in_port("alice"))
     alice_procnode.node.entdist_in_port.connect(entdistcomp.node_out_port("alice"))
     bob_procnode.node.entdist_out_port.connect(entdistcomp.node_in_port("bob"))
     bob_procnode.node.entdist_in_port.connect(entdistcomp.node_out_port("bob"))
-    entdist = EntDist(nodes=nodes, network_info=network_info, comp=entdistcomp)
+    entdist = EntDist(nodes=nodes, ehi_network=network_ehi, comp=entdistcomp)
     link_info = LhiLinkInfo.perfect(1000)
     entdist.add_sampler(alice_procnode.node.ID, bob_procnode.node.ID, link_info)
 
+    # To prevent scheduler from running
+    alice_procnode.scheduler = MockScheduler(alice_procnode.scheduler)
+    bob_procnode.scheduler = MockScheduler(bob_procnode.scheduler)
+
     # First start Bob, since Alice won't yield on anything (she only does a Send
     # instruction) and therefore calling 'start()' on alice completes her whole
     # protocol while Bob's interface has not even been started.
@@ -810,7 +783,7 @@ META_END
 ^b0 {type = CL}:
     remote_id = assign_cval() : {client_id}
 ^b1 {type = QC}:
-    run_request(vec<>) : req1
+    run_request() : req1
 
 REQUEST req1
   callback_type: wait_all
@@ -830,12 +803,12 @@ REQUEST req1
     num_qubits = 3
     topology = generic_topology(num_qubits)
     latencies = LhiLatencies.all_zero()
-    ntf = GenericToVanillaInterface()
+    ntf = GenericNtf()
     ehi = LhiConverter.to_ehi(topology, ntf, latencies)
     unit_module = UnitModule.from_full_ehi(ehi)
-    network_info = create_network_info(num_qubits, names=["client", "server"])
-    server_id = network_info.get_node_id("server")
-    client_id = network_info.get_node_id("client")
+    network_ehi = create_network_ehi(num_qubits, names=["client", "server"])
+    server_id = network_ehi.get_node_id("server")
+    client_id = network_ehi.get_node_id("client")
 
     class TestProcNode(ProcNode):
         def run(self) -> Generator[EventExpression, None, None]:
@@ -846,7 +819,7 @@ REQUEST req1
 
     server_procnode = create_procnode(
         "server",
-        network_info,
+        network_ehi,
         num_qubits,
         topology,
         latencies,
@@ -873,7 +846,7 @@ META_END
 ^b0 {type = CL}:
     remote_id = assign_cval() : {server_id}
 ^b1 {type = QL}:
-    run_request(vec<>) : req1
+    run_request() : req1
 
 REQUEST req1
   callback_type: wait_all
@@ -891,7 +864,7 @@ REQUEST req1
     client_program = QoalaParser(client_text).parse()
     client_procnode = create_procnode(
         "client",
-        network_info,
+        network_ehi,
         num_qubits,
         topology,
         latencies,
@@ -910,15 +883,19 @@ REQUEST req1
     client_procnode.connect_to(server_procnode)
 
     nodes = [client_procnode.node, server_procnode.node]
-    entdistcomp = EntDistComponent(network_info)
+    entdistcomp = EntDistComponent(network_ehi)
     client_procnode.node.entdist_out_port.connect(entdistcomp.node_in_port("client"))
     client_procnode.node.entdist_in_port.connect(entdistcomp.node_out_port("client"))
     server_procnode.node.entdist_out_port.connect(entdistcomp.node_in_port("server"))
     server_procnode.node.entdist_in_port.connect(entdistcomp.node_out_port("server"))
-    entdist = EntDist(nodes=nodes, network_info=network_info, comp=entdistcomp)
+    entdist = EntDist(nodes=nodes, ehi_network=network_ehi, comp=entdistcomp)
     link_info = LhiLinkInfo.perfect(1000)
     entdist.add_sampler(client_procnode.node.ID, server_procnode.node.ID, link_info)
 
+    # To prevent scheduler from running
+    server_procnode.scheduler = MockScheduler(server_procnode.scheduler)
+    client_procnode.scheduler = MockScheduler(client_procnode.scheduler)
+
     server_procnode.start()
     client_procnode.start()
     entdist.start()
@@ -936,7 +913,6 @@ REQUEST req1
 
 if __name__ == "__main__":
     test_initialize()
-    test_2()
     test_classical_comm()
     test_classical_comm_three_nodes()
     test_epr()
