@@ -1,64 +1,101 @@
 import netsquid as ns
 import pytest
 from netsquid.components import instructions as ns_instr
+from netsquid.components.instructions import (
+    INSTR_CNOT,
+    INSTR_CXDIR,
+    INSTR_CYDIR,
+    INSTR_H,
+    INSTR_INIT,
+    INSTR_MEASURE,
+    INSTR_ROT_X,
+    INSTR_ROT_Y,
+    INSTR_ROT_Z,
+    INSTR_X,
+    INSTR_Y,
+    INSTR_Z,
+)
 from netsquid.components.qprocessor import MissingInstructionError
 from netsquid.nodes import Node
 from netsquid.qubits import ketstates
 
-from qoala.runtime.config import GenericQDeviceConfig, NVQDeviceConfig
-from qoala.sim.build import build_generic_qprocessor, build_nv_qprocessor
-from qoala.sim.constants import PI, PI_OVER_2
+from qoala.runtime.lhi import LhiTopologyBuilder
+from qoala.sim.build import build_qprocessor_from_topology
 from qoala.sim.qdevice import (
-    GenericPhysicalQuantumMemory,
     NonInitializedQubitError,
-    NVPhysicalQuantumMemory,
     QDevice,
     QDeviceCommand,
-    QDeviceType,
     UnsupportedQDeviceCommandError,
 )
-from qoala.util.tests import has_state, netsquid_run
+from qoala.util.math import PI, PI_OVER_2, has_state
+from qoala.util.tests import netsquid_run
 
 
-def perfect_generic_qdevice(num_qubits: int) -> QDevice:
-    cfg = GenericQDeviceConfig.perfect_config(num_qubits=num_qubits)
-    processor = build_generic_qprocessor(name="processor", cfg=cfg)
-    node = Node(name="alice", qmemory=processor)
-    return QDevice(
-        node=node,
-        typ=QDeviceType.GENERIC,
-        memory=GenericPhysicalQuantumMemory(num_qubits),
+def perfect_uniform_qdevice(num_qubits: int) -> QDevice:
+    topology = LhiTopologyBuilder.perfect_uniform(
+        num_qubits=num_qubits,
+        single_instructions=[
+            INSTR_INIT,
+            INSTR_X,
+            INSTR_Y,
+            INSTR_Z,
+            INSTR_H,
+            INSTR_ROT_X,
+            INSTR_ROT_Y,
+            INSTR_ROT_Z,
+            INSTR_MEASURE,
+        ],
+        single_duration=5e3,
+        two_instructions=[INSTR_CNOT],
+        two_duration=100e3,
     )
-
-
-def perfect_nv_qdevice(num_qubits: int) -> QDevice:
-    cfg = NVQDeviceConfig.perfect_config(num_qubits=num_qubits)
-    processor = build_nv_qprocessor(name="processor", cfg=cfg)
+    processor = build_qprocessor_from_topology(name="processor", topology=topology)
     node = Node(name="alice", qmemory=processor)
-    return QDevice(
-        node=node,
-        typ=QDeviceType.NV,
-        memory=NVPhysicalQuantumMemory(num_qubits),
+    return QDevice(node=node, topology=topology)
+
+
+def perfect_nv_star_qdevice(num_qubits: int) -> QDevice:
+    topology = LhiTopologyBuilder.perfect_star(
+        num_qubits=num_qubits,
+        comm_instructions=[
+            INSTR_INIT,
+            INSTR_ROT_X,
+            INSTR_ROT_Y,
+            INSTR_ROT_Z,
+            INSTR_MEASURE,
+        ],
+        comm_duration=5e3,
+        mem_instructions=[
+            INSTR_INIT,
+            INSTR_ROT_X,
+            INSTR_ROT_Y,
+            INSTR_ROT_Z,
+        ],
+        mem_duration=1e4,
+        two_instructions=[INSTR_CXDIR, INSTR_CYDIR],
+        two_duration=1e5,
     )
+    processor = build_qprocessor_from_topology(name="processor", topology=topology)
+    node = Node(name="alice", qmemory=processor)
+    return QDevice(node=node, topology=topology)
 
 
 def test_static_generic():
     num_qubits = 3
-    qdevice = perfect_generic_qdevice(num_qubits)
+    qdevice = perfect_uniform_qdevice(num_qubits)
 
     assert qdevice.qprocessor.num_positions == num_qubits
 
-    assert qdevice.typ == QDeviceType.GENERIC
-    assert qdevice.qubit_count == num_qubits
-    assert qdevice.comm_qubit_count == num_qubits
-    assert qdevice.comm_qubit_ids == {i for i in range(num_qubits)}
-    assert qdevice.mem_qubit_ids == {i for i in range(num_qubits)}
-    assert qdevice.all_qubit_ids == {i for i in range(num_qubits)}
+    assert qdevice.get_qubit_count() == num_qubits
+    assert qdevice.get_comm_qubit_count() == num_qubits
+    assert qdevice.get_comm_qubit_ids() == {i for i in range(num_qubits)}
+    assert qdevice.get_non_comm_qubit_ids() == set()
+    assert qdevice.get_all_qubit_ids() == {i for i in range(num_qubits)}
 
 
 def test_static_nv():
     num_qubits = 3
-    qdevice = perfect_nv_qdevice(num_qubits)
+    qdevice = perfect_nv_star_qdevice(num_qubits)
 
     assert qdevice.qprocessor.num_positions == num_qubits
     with pytest.raises(MissingInstructionError):
@@ -68,17 +105,18 @@ def test_static_nv():
     # Should not raise error:
     qdevice.qprocessor.get_instruction_duration(ns_instr.INSTR_CXDIR, [0, 1])
 
-    assert qdevice.typ == QDeviceType.NV
-    assert qdevice.qubit_count == num_qubits
-    assert qdevice.comm_qubit_count == 1
-    assert qdevice.comm_qubit_ids == {0}
-    assert qdevice.mem_qubit_ids == {i for i in range(1, num_qubits)}
-    assert qdevice.all_qubit_ids == {i for i in range(num_qubits)}
+    assert qdevice.get_qubit_count() == num_qubits
+    assert qdevice.get_comm_qubit_count() == 1
+    assert qdevice.get_comm_qubit_ids() == {0}
+    assert qdevice.get_non_comm_qubit_ids() == {i for i in range(1, num_qubits)}
+    assert qdevice.get_all_qubit_ids() == {i for i in range(num_qubits)}
 
 
 def test_initalize_generic():
+    ns.sim_reset()
+
     num_qubits = 3
-    qdevice = perfect_generic_qdevice(num_qubits)
+    qdevice = perfect_uniform_qdevice(num_qubits)
 
     # All qubits are not initalized yet.
     assert qdevice.get_local_qubit(0) is None
@@ -125,8 +163,10 @@ def test_initalize_generic():
 
 
 def test_initalize_nv():
+    ns.sim_reset()
+
     num_qubits = 3
-    qdevice = perfect_nv_qdevice(num_qubits)
+    qdevice = perfect_nv_star_qdevice(num_qubits)
 
     # All qubits are not initalized yet.
     assert qdevice.get_local_qubit(0) is None
@@ -183,8 +223,10 @@ def test_initalize_nv():
 
 
 def test_rotations_generic():
+    ns.sim_reset()
+
     num_qubits = 3
-    qdevice = perfect_generic_qdevice(num_qubits)
+    qdevice = perfect_uniform_qdevice(num_qubits)
 
     # All qubits are not initalized yet.
     assert qdevice.get_local_qubit(0) is None
@@ -227,8 +269,10 @@ def test_rotations_generic():
 
 
 def test_rotations_nv():
+    ns.sim_reset()
+
     num_qubits = 3
-    qdevice = perfect_nv_qdevice(num_qubits)
+    qdevice = perfect_nv_star_qdevice(num_qubits)
 
     # All qubits are not initalized yet.
     assert qdevice.get_local_qubit(0) is None
@@ -281,7 +325,7 @@ def test_rotations_nv():
 
 def test_measure_generic():
     num_qubits = 3
-    qdevice = perfect_generic_qdevice(num_qubits)
+    qdevice = perfect_uniform_qdevice(num_qubits)
 
     # Initialize qubit 0 and measure it.
     commands = [
@@ -335,7 +379,7 @@ def test_measure_generic():
 
 def test_measure_nv():
     num_qubits = 3
-    qdevice = perfect_nv_qdevice(num_qubits)
+    qdevice = perfect_nv_star_qdevice(num_qubits)
 
     # Initialize qubit 0 and measure it.
     commands = [
@@ -376,8 +420,10 @@ def test_measure_nv():
 
 
 def test_two_qubit_gates_generic():
+    ns.sim_reset()
+
     num_qubits = 3
-    qdevice = perfect_generic_qdevice(num_qubits)
+    qdevice = perfect_uniform_qdevice(num_qubits)
 
     # All qubits are not initalized yet.
     assert qdevice.get_local_qubit(0) is None
@@ -413,8 +459,10 @@ def test_two_qubit_gates_generic():
 
 
 def test_two_qubit_gates_nv():
+    ns.sim_reset()
+
     num_qubits = 3
-    qdevice = perfect_nv_qdevice(num_qubits)
+    qdevice = perfect_nv_star_qdevice(num_qubits)
 
     # All qubits are not initalized yet.
     assert qdevice.get_local_qubit(0) is None
@@ -453,7 +501,7 @@ def test_two_qubit_gates_nv():
 
 def test_unsupported_commands_generic():
     num_qubits = 3
-    qdevice = perfect_generic_qdevice(num_qubits)
+    qdevice = perfect_uniform_qdevice(num_qubits)
 
     with pytest.raises(UnsupportedQDeviceCommandError):
         commands = [QDeviceCommand(ns_instr.INSTR_T, [0])]
@@ -462,7 +510,7 @@ def test_unsupported_commands_generic():
 
 def test_unsupported_commands_nv():
     num_qubits = 3
-    qdevice = perfect_nv_qdevice(num_qubits)
+    qdevice = perfect_nv_star_qdevice(num_qubits)
 
     with pytest.raises(UnsupportedQDeviceCommandError):
         commands = [QDeviceCommand(ns_instr.INSTR_X, [0])]
@@ -507,7 +555,7 @@ def test_unsupported_commands_nv():
 
 def test_non_initalized():
     num_qubits = 3
-    qdevice = perfect_generic_qdevice(num_qubits)
+    qdevice = perfect_uniform_qdevice(num_qubits)
 
     with pytest.raises(NonInitializedQubitError):
         commands = [QDeviceCommand(ns_instr.INSTR_X, [0])]
