@@ -8,6 +8,8 @@ from qoala.lang.hostlang import (
     AssignCValueOp,
     BasicBlock,
     BasicBlockType,
+    IqoalaSingleton,
+    IqoalaTuple,
     IqoalaVector,
     ReceiveCMsgOp,
     ReturnResultOp,
@@ -62,7 +64,8 @@ def test_serialize_host_code_1():
     new_value = assign_cval() : 3
     my_value = add_cval_c(new_value, new_value)
 ^b1 {type = QL}:
-    vec<m> = run_subroutine(vec<my_value>) : subrt1
+    tuple<m> = run_subroutine(tuple<my_value>) : subrt1
+    x<5> = run_subroutine() : subrt2
 ^b2 {type = CL}:
     return_result(m)
     """
@@ -71,30 +74,38 @@ def test_serialize_host_code_1():
         "b0",
         BasicBlockType.CL,
         instructions=[
-            AssignCValueOp("my_value", 1),
-            AssignCValueOp("remote_id", 0),
-            SendCMsgOp("remote_id", "my_value"),
-            ReceiveCMsgOp("remote_id", "received_value"),
-            AssignCValueOp("new_value", 3),
-            AddCValueOp("my_value", "new_value", "new_value"),
+            AssignCValueOp(IqoalaSingleton("my_value"), 1),
+            AssignCValueOp(IqoalaSingleton("remote_id"), 0),
+            SendCMsgOp(IqoalaSingleton("remote_id"), IqoalaSingleton("my_value")),
+            ReceiveCMsgOp(
+                IqoalaSingleton("remote_id"), IqoalaSingleton("received_value")
+            ),
+            AssignCValueOp(IqoalaSingleton("new_value"), 3),
+            AddCValueOp(
+                IqoalaSingleton("my_value"),
+                IqoalaSingleton("new_value"),
+                IqoalaSingleton("new_value"),
+            ),
         ],
     )
     b1 = BasicBlock(
         "b1",
         BasicBlockType.QL,
         instructions=[
-            RunSubroutineOp(IqoalaVector(["m"]), IqoalaVector(["my_value"]), "subrt1"),
+            RunSubroutineOp(IqoalaTuple(["m"]), IqoalaTuple(["my_value"]), "subrt1"),
+            RunSubroutineOp(IqoalaVector("x", 5), IqoalaTuple([]), "subrt2"),
         ],
     )
     b2 = BasicBlock(
         "b2",
         BasicBlockType.CL,
         instructions=[
-            ReturnResultOp("m"),
+            ReturnResultOp(IqoalaSingleton("m")),
         ],
     )
 
     program = QoalaProgram(meta=ProgramMeta.empty("alice"), blocks=[b0, b1, b2])
+    print(program.serialize_host_code())
     assert text_equal(program.serialize_host_code(), expected)
 
 
@@ -117,7 +128,7 @@ SUBROUTINE subrt1
     M0 = Register.from_str("M0")
     subrt = LocalRoutine(
         name="subrt1",
-        subrt=Subroutine(
+        subroutine=Subroutine(
             instructions=[
                 SetInstruction(reg=Q0, imm=0),
                 RotZInstruction(reg=Q0, imm0=Template("my_value"), imm1=4),
@@ -141,7 +152,7 @@ def test_serialize_subroutines_2():
     expected = """
 SUBROUTINE subrt1
     params: param1
-    returns: m
+    returns: outcomes<10>
     uses: 0
     keeps: 
   NETQASM_START
@@ -164,19 +175,19 @@ SUBROUTINE subrt2
     M0 = Register.from_str("M0")
     subrt1 = LocalRoutine(
         name="subrt1",
-        subrt=Subroutine(
+        subroutine=Subroutine(
             instructions=[
                 SetInstruction(reg=R0, imm=Template("param1")),
                 MeasInstruction(reg0=Q0, reg1=M0),
             ],
             arguments=["param1"],
         ),
-        return_vars=["m"],
+        return_vars=[IqoalaVector("outcomes", 10)],
         metadata=RoutineMetadata.free_all([0]),
     )
     subrt2 = LocalRoutine(
         name="subrt2",
-        subrt=Subroutine(
+        subroutine=Subroutine(
             instructions=[
                 SetInstruction(reg=R0, imm=Template("theta")),
             ],
@@ -210,7 +221,7 @@ META_END
     new_value = assign_cval() : 3
     my_value = add_cval_c(new_value, new_value)
 ^b1 {type = QL}:
-    vec<m> = run_subroutine(vec<my_value>) : subrt1
+    tuple<m> = run_subroutine(tuple<my_value>) : subrt1
 ^b2 {type = CL}:
     return_result(m)
 
@@ -237,7 +248,7 @@ SUBROUTINE subrt1
         AddCValueOp("my_value", "new_value", "new_value"),
     ]
     b1_instructions = [
-        RunSubroutineOp(IqoalaVector(["m"]), IqoalaVector(["my_value"]), "subrt1")
+        RunSubroutineOp(IqoalaTuple(["m"]), IqoalaTuple(["my_value"]), "subrt1")
     ]
     b2_instructions = [ReturnResultOp("m")]
     b0 = BasicBlock("b0", BasicBlockType.CL, b0_instructions)
@@ -247,7 +258,7 @@ SUBROUTINE subrt1
     M0 = Register.from_str("M0")
     subrt = LocalRoutine(
         name="subrt1",
-        subrt=Subroutine(
+        subroutine=Subroutine(
             instructions=[
                 SetInstruction(reg=Q0, imm=0),
                 RotZInstruction(reg=Q0, imm0=Template("my_value"), imm1=4),
